# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/schema__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from tecton_proto.common import column_type__client_pb2 as tecton__proto_dot_common_dot_column__type____client__pb2
from tecton_proto.common import data_type__client_pb2 as tecton__proto_dot_common_dot_data__type____client__pb2
from tecton_proto.common import time_window__client_pb2 as tecton__proto_dot_common_dot_time__window____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(tecton_proto/common/schema__client.proto\x12\x13tecton_proto.common\x1a\x1egoogle/protobuf/duration.proto\x1a-tecton_proto/common/column_type__client.proto\x1a+tecton_proto/common/data_type__client.proto\x1a-tecton_proto/common/time_window__client.proto\"\xe4\x02\n\x06\x43olumn\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12I\n\x11offline_data_type\x18\x05 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x0fofflineDataType\x12V\n\x18\x66\x65\x61ture_server_data_type\x18\x06 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x15\x66\x65\x61tureServerDataType\x12$\n\x0eraw_spark_type\x18\x02 \x01(\tR\x0crawSparkType\x12,\n\x12raw_snowflake_type\x18\x04 \x01(\tR\x10rawSnowflakeType\x12O\n\x13\x66\x65\x61ture_server_type\x18\x03 \x01(\x0e\x32\x1f.tecton_proto.common.ColumnTypeR\x11\x66\x65\x61tureServerType\"E\n\x06Schema\x12\x35\n\x07\x63olumns\x18\x02 \x03(\x0b\x32\x1b.tecton_proto.common.ColumnR\x07\x63olumnsJ\x04\x08\x01\x10\x02\"|\n\x16OnlineBatchTableFormat\x12\x62\n\x18online_batch_table_parts\x18\x03 \x03(\x0b\x32).tecton_proto.common.OnlineBatchTablePartR\x15onlineBatchTableParts\"\xf5\x01\n\x14OnlineBatchTablePart\x12!\n\x0cwindow_index\x18\x01 \x01(\x05R\x0bwindowIndex\x12@\n\x0btime_window\x18\x02 \x01(\x0b\x32\x1f.tecton_proto.common.TimeWindowR\ntimeWindow\x12\x33\n\x06schema\x18\x03 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x06schema\x12\x43\n\x05tiles\x18\x04 \x03(\x0b\x32-.tecton_proto.common.OnlineBatchTablePartTileR\x05tiles\"\xd2\x01\n\x18OnlineBatchTablePartTile\x12\\\n\x1drelative_start_time_inclusive\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x1arelativeStartTimeInclusive\x12X\n\x1brelative_end_time_exclusive\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x18relativeEndTimeExclusiveB\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.schema__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _COLUMN._serialized_start=237
  _COLUMN._serialized_end=593
  _SCHEMA._serialized_start=595
  _SCHEMA._serialized_end=664
  _ONLINEBATCHTABLEFORMAT._serialized_start=666
  _ONLINEBATCHTABLEFORMAT._serialized_end=790
  _ONLINEBATCHTABLEPART._serialized_start=793
  _ONLINEBATCHTABLEPART._serialized_end=1038
  _ONLINEBATCHTABLEPARTTILE._serialized_start=1041
  _ONLINEBATCHTABLEPARTTILE._serialized_end=1251
# @@protoc_insertion_point(module_scope)



from tecton_core.query import node_interface
from tecton_core.query import nodes
from tecton_core.query import rewrite
from tecton_snowflake.query import nodes as snowflake_nodes


class SnowflakeRewrite(rewrite.Rewrite):
    """
    Replace with Snowflake specific node.
    """

    node_mapping = {
        nodes.PartialAggNode: snowflake_nodes.PartialAggSnowflakeNode,
        nodes.AsofJoinFullAggNode: snowflake_nodes.AsofJoinFullAggSnowflakeNode,
    }

    def rewrite(self, tree: node_interface.NodeRef) -> node_interface.NodeRef:
        for i in tree.inputs:
            self.rewrite(i)
        if tree.node.__class__ in self.node_mapping:
            tree.node = self.node_mapping[tree.node.__class__].from_query_node(tree.node)



import urllib.parse
from typing import Dict

from tecton import version
from tecton_core import conf
from tecton_core import errors
from tecton_core.id_helper import IdHelper
from tecton_core.metadata_service_impl.providers import RequestProvider
from tecton_core.metadata_service_impl.trace import get_trace_id


class InternalRequestProvider(RequestProvider):
    def request_headers(self) -> Dict[str, str]:
        """
        :return: Dictionary of request metadata.
        """
        metadata = {}

        metadata["x-request-id"] = IdHelper.generate_string_id()
        trace_id = get_trace_id()
        if trace_id:
            metadata["x-trace-id"] = trace_id

        metadata["x-tecton-client-version"] = version.get_semantic_version()

        workspace = conf.get_or_none("TECTON_WORKSPACE")
        if workspace:
            metadata["x-workspace"] = workspace
            # Warning: This is a hack to make it possible to integration test both EMR and Databricks
            # in a single deployment.
            if workspace.endswith("__emr"):
                metadata["x-tecton-force-emr"] = "true"

        authorization = self.auth_provider.get_auth_header()
        if authorization:
            metadata["authorization"] = authorization

        parsed_url = urllib.parse.urlparse(self.request_url())
        metadata["host"] = parsed_url.netloc

        return metadata

    def request_url(self) -> str:
        """
        :return: A Validated API service URL.
        """
        api_service = conf.get_or_none("API_SERVICE")
        if not api_service:
            msg = "API_SERVICE not set. Please configure API_SERVICE or use tecton.set_credentials(tecton_url=<url>)"
            raise errors.TectonAPIValidationError(msg)
        conf.validate_api_service_url(api_service)
        return urllib.parse.urljoin(api_service + "/", "proxy")



from __future__ import annotations

import random
import types
from dataclasses import dataclass
from logging import getLogger
from typing import Any
from typing import Iterator
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple

import numpy.random
import pyarrow
import pyarrow.compute as pc
import torch
import torch.nn.functional as F
from numpy import typing as np_typing
from transformers import AutoModel
from transformers import AutoTokenizer
from transformers import BatchEncoding
from transformers import PretrainedConfig
from transformers import tokenization_utils_base

from tecton_core.embeddings import execution_utils


logger = getLogger(__name__)

_MAX_INFERENCE_ATTEMPTS = 2
# Fraction of GPU memory to use for the token budget estimation.
_MAX_GPU_ALLOCATION = 0.3
_DEFAULT_GPU_TOKEN_BUDGET = 10_000
_DEFAULT_CPU_TOKEN_BUDGET = 1_000
_MAX_TOKENIZATION_LENGTH = 512
_DEFAULT_TORCH_DTYPE = torch.float32
_INPUT_IDS_COL_NAME = "_tecton_internal_input_ids"
_TOKEN_TYPE_IDS_COL_NAME = "_tecton_internal_token_type_ids"
_ATTENTION_MASK_COL_NAME = "_tecton_internal_attention_mask"


@dataclass
class InferenceParameters:
    max_input_size: int
    output_size: int
    torch_dtype: torch.dtype
    model_type: str
    _config: PretrainedConfig

    @classmethod
    def from_config(cls, config: PretrainedConfig) -> InferenceParameters:
        return InferenceParameters(
            max_input_size=config.max_position_embeddings or _MAX_TOKENIZATION_LENGTH,
            torch_dtype=config.torch_dtype or _DEFAULT_TORCH_DTYPE,
            output_size=config.hidden_size,
            model_type=config.model_type,
            _config=config,
        )


@dataclass
class ModelContainer:
    tokenizer: tokenization_utils_base.PreTrainedTokenizerBase
    model: torch.nn.Module
    config: InferenceParameters

    @classmethod
    def load(cls, filepath: str, device_name: str) -> ModelContainer:
        with torch.device(device_name):
            model = AutoModel.from_pretrained(
                filepath,
                use_safetensors=True,
                use_cache=False,
                local_files_only=True,
                trust_remote_code=False,
            )
            return cls(
                tokenizer=AutoTokenizer.from_pretrained(filepath, trust_remote_code=False),
                model=model.eval(),
                config=InferenceParameters.from_config(model.config),
            )


# Follows pattern from: https://huggingface.co/sentence-transformers/all-MiniLM-L6-v2
def _mean_pooling(*, token_embeddings: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
    input_mask_expanded = attention_mask.unsqueeze(dim=-1).expand(token_embeddings.size()).to(token_embeddings.dtype)
    sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, dim=1)
    sum_mask = torch.sum(input_mask_expanded, dim=1)
    # NOTE: clamps to 1e-9 to avoid NaN from the division
    return sum_embeddings / torch.clamp(sum_mask, min=1e-9)


def _embed_pytorch_core_logic(
    *,
    encoded_input: Mapping[str, torch.Tensor],
    model: torch.nn.Module,
) -> torch.Tensor:
    with torch.no_grad():
        model_output = model(**encoded_input)

    # TODO: encapsulate this as its own nn.Module so we can have a single nn.Module representation of our model.
    token_embeddings = model_output[0]  # First element of model_output contains all token embeddings
    sentence_embeddings = _mean_pooling(
        token_embeddings=token_embeddings, attention_mask=encoded_input["attention_mask"]
    )
    sentence_embeddings = F.normalize(sentence_embeddings, p=2, dim=1)
    return sentence_embeddings


def _embed_pytorch_core_logic_with_bisect(
    *,
    encoded_input: Mapping[str, torch.Tensor],
    model: torch.nn.Module,
) -> torch.Tensor:
    # All tensors should have the same size at this point.
    mid_point = len(encoded_input[_INPUT_IDS_COL_NAME].size()) // 2

    first_chunk, second_chunk = {}, {}

    for key, value in encoded_input.items():
        first_half, second_half = value[:mid_point], value[mid_point:]
        first_chunk[key] = first_half
        second_chunk[key] = second_half

    return torch.cat(
        [
            _embed_pytorch_core_logic(encoded_input=first_chunk, model=model),
            _embed_pytorch_core_logic(encoded_input=second_chunk, model=model),
        ]
    )


def _embed_pytorch_core_logic_with_retries(
    *,
    encoded_input: Mapping[str, torch.Tensor],
    model: torch.nn.Module,
    max_attempts: int = _MAX_INFERENCE_ATTEMPTS,
) -> torch.Tensor:
    while max_attempts > 0:
        max_attempts -= 1
        try:
            return _embed_pytorch_core_logic(encoded_input=encoded_input, model=model)
        except torch.cuda.OutOfMemoryError as oom_exc:
            logger.warning(torch.cuda.memory_summary())
            logger.warning(f"Out of Memory. {oom_exc}\n Trying again {max_attempts} more time(s).")
            execution_utils.gc_all(torch.cuda.current_device())
            if max_attempts == 0:
                break
    logger.warning("Failed to embed strings after multiple attempts. Bisecting batch and trying again.")
    return _embed_pytorch_core_logic_with_bisect(encoded_input=encoded_input, model=model)


def _convert_pytorch_to_arrow(sentence_embeddings: torch.Tensor, num_nulls_end: int = 0) -> pyarrow.Array:
    if len(sentence_embeddings.shape) != 2:
        msg = f"Cannot convert torch.Tensor of shape: {sentence_embeddings.shape}, must be a 2d tensor."
        raise ValueError(msg)
    if sentence_embeddings.dtype != torch.float32:
        msg = f"Cannot convert torch.Tensor of dtype: {sentence_embeddings.dtype}, must be torch.float32."
        raise ValueError(msg)

    sentence_embeddings = sentence_embeddings.cpu().numpy()

    # NOTE: we need to convert our sentence_embeddings to a list and pad it with any nulls.
    sentence_embeddings = [*sentence_embeddings, *(None for _ in range(num_nulls_end))]

    # NOTE: using dynamic sized pyarrow list due to https://github.com/apache/arrow/issues/35697
    return pyarrow.array(list(sentence_embeddings), type=pyarrow.list_(pyarrow.float32()))


def _random_string(tokenizer: tokenization_utils_base.PreTrainedTokenizerBase, max_tokens_in_string: int) -> str:
    num_tokens = numpy.random.randint(3, max_tokens_in_string)
    random_token_ids = numpy.random.randint(0, tokenizer.vocab_size, size=num_tokens)
    tokens = tokenizer.convert_ids_to_tokens(random_token_ids)
    return tokenizer.convert_tokens_to_string(tokens)


def _record_batch_from_vocab(
    tokenizer: tokenization_utils_base.PreTrainedTokenizerBase, tokens_per_line: int, rows: int, column_name: str
) -> pyarrow.RecordBatch:
    text_array = []
    vocab = list(tokenizer.get_vocab().keys())
    # TODO replace below with _random_string once we use tokenization before batching
    for i in range(rows):
        text_array.append(
            {column_name: " ".join([random.choice(vocab) for _ in range(random.randint(3, tokens_per_line))])}
        )
    schema = pyarrow.schema([pyarrow.field(column_name, pyarrow.string())])
    return pyarrow.RecordBatch.from_pylist(text_array, schema=schema)


def _embed_pytorch(
    *,
    batch: pyarrow.RecordBatch,
    model: torch.nn.Module,
    input_column_name: str,
    output_column_name: str,
) -> pyarrow.RecordBatch:
    input_column = batch.column(input_column_name)
    null_index = pc.index(pc.is_null(input_column), pyarrow.scalar(True)).as_py()

    if null_index == 0:
        # NOTE: this case means that they are all null! Short circuit and create the embeddings directly.
        sentence_embeddings = pyarrow.nulls(len(input_column), pyarrow.list_(pyarrow.float32()))
    elif null_index > -1:
        # No batch contains null value after non-null value.
        # The rule is last microbatch always only has null values. Any microbatch before the last one should not have any null value.
        msg = "Unexpected non-null value in the null value microbatch"
        raise ValueError(msg)
    else:
        input_ids = torch.tensor(batch.column(_INPUT_IDS_COL_NAME).to_pylist())
        token_type_ids = torch.tensor(batch.column(_TOKEN_TYPE_IDS_COL_NAME).to_pylist())
        attention_mask = torch.tensor(batch.column(_ATTENTION_MASK_COL_NAME).to_pylist())

        encoded_input = {"input_ids": input_ids, "token_type_ids": token_type_ids, "attention_mask": attention_mask}

        sentence_embeddings = _embed_pytorch_core_logic_with_retries(encoded_input=encoded_input, model=model)

        sentence_embeddings = _convert_pytorch_to_arrow(sentence_embeddings)

    columns_to_drop = [_INPUT_IDS_COL_NAME, _TOKEN_TYPE_IDS_COL_NAME, _ATTENTION_MASK_COL_NAME]
    remaining_columns = [batch.column(col_name) for col_name in batch.schema.names if col_name not in columns_to_drop]
    remaing_schema = pyarrow.schema(
        [batch.schema.field(col_name) for col_name in batch.schema.names if col_name not in columns_to_drop]
    )
    batch_with_results = pyarrow.RecordBatch.from_arrays(
        [*remaining_columns, sentence_embeddings],
        schema=remaing_schema.append(pyarrow.field(output_column_name, sentence_embeddings.type)),
    )
    return batch_with_results


def _token_budget_from_tokens(tokens: BatchEncoding) -> int:
    total_tokens = 0
    for i in range(len(tokens.items())):
        total_tokens += len(tokens.tokens(i))
    return total_tokens


def embed_pytorch(
    batch: pyarrow.RecordBatch,
    *,
    model: ModelContainer,
    input_column_name: str,
    output_column_name: str,
    cuda_device: str,
) -> pyarrow.RecordBatch:
    with torch.device(cuda_device):
        results = _embed_pytorch(
            batch=batch,
            model=model.model,
            input_column_name=input_column_name,
            output_column_name=output_column_name,
        )
    return results


def _estimate_token_budget_for_gpu(
    inference_info: Tuple[execution_utils.ModelCallable, Sequence[execution_utils.FuncConfig]],
) -> int:
    """
    Estimate the token budget for a given model on a GPU by running inference on a small batch of data that is
    generated from the token vocabulary in the model's vocabulary. The amount of memory used on the GPU for that small
    batch is then used to infer the amount of memory that would be used by a larger batch and worked back into a
    token budget.

    This method is not perfect and has some limitations:
    * The method to estimate the number of tokens from a string's length get more inaccurate as the strings gets shorter,
    due small strings having a less consistent number of tokens per character.
    * A small increase in the number of tokens can dramatically increase the size of the array due to padding, thereby
      increasing the memory usage.
    * GPU_MEMORY_FRACTION is set to a lower number since the number of tokens per line is determined by the line that
      expands to the most tokens.
    """
    inference_func, inference_config = inference_info
    inference_kwargs = inference_config[0].load().kwargs()
    model_container: ModelContainer = inference_kwargs["model"]
    input_column_name = inference_kwargs["input_column_name"]
    batches_to_estimate = 3
    synthetic_batch = _record_batch_from_vocab(model_container.tokenizer, 300, 10000, input_column_name)
    batches = list(
        dynamic_token_batching(
            synthetic_batch,
            token_budget=_DEFAULT_GPU_TOKEN_BUDGET,
            column_name=input_column_name,
            tokenizer=model_container.tokenizer,
            max_input_size=model_container.config.max_input_size,
        )
    )
    print("Estimating token budget for GPU.")
    with torch.device("cuda:0"):
        budgets = []
        for dynamic_batch in batches[:batches_to_estimate]:
            memory_tracker = execution_utils.CudaMemoryTracker("cuda:0")
            with memory_tracker:
                _embed_pytorch(
                    batch=dynamic_batch,
                    model=inference_kwargs["model"].model,
                    input_column_name=inference_kwargs["input_column_name"],
                    output_column_name=inference_kwargs["output_column_name"],
                )
            # See doc string for explanation of the formula
            memory_usage = memory_tracker.get_usage_data()
            max_token_budget = memory_usage.max_token_budget_from_mem_alloc(_DEFAULT_GPU_TOKEN_BUDGET)
            estimated_safe_token_budget = int(max_token_budget * _MAX_GPU_ALLOCATION)
            budgets.append(estimated_safe_token_budget)
        print(f"Estimated token budgets: {budgets}, using {min(budgets)} as the budget.")
        return min(budgets)


def estimate_token_budget(
    inference: Tuple[execution_utils.ModelCallable, Sequence[execution_utils.FuncConfig]],
    device_type: str,
) -> int:
    if device_type == "cpu":
        print(
            f"No estimation method implemented for CPU, using default CPU token budget of {_DEFAULT_CPU_TOKEN_BUDGET}."
        )
        return _DEFAULT_CPU_TOKEN_BUDGET
    else:
        try:
            token_budget = _estimate_token_budget_for_gpu(inference)
            if token_budget <= 0:
                msg = "Token budget estimation returned a negative value."
                raise ValueError(msg)
            return token_budget
        except Exception as e:
            logger.warning(f"Failed to estimate token budget for GPU: {e}")
            return _DEFAULT_GPU_TOKEN_BUDGET


@dataclass
class EmbedPytorchFuncConfig:
    _model_filename: str
    _cuda_device: str
    _extra_kwargs: Mapping[str, Any]

    # Note: _final_kwargs is a late init object.
    _final_kwargs: Optional[Mapping[str, Any]] = None

    @classmethod
    def create(
        cls, model_filename: str, cuda_device: str, input_column_name: str, output_column_name: str
    ) -> EmbedPytorchFuncConfig:
        return cls(
            _model_filename=model_filename,
            _cuda_device=cuda_device,
            _extra_kwargs=types.MappingProxyType(
                {
                    "cuda_device": cuda_device,
                    "input_column_name": input_column_name,
                    "output_column_name": output_column_name,
                }
            ),
        )

    def load(self):
        model_container = ModelContainer.load(self._model_filename, device_name=self._cuda_device)
        self._final_kwargs = types.MappingProxyType(dict(model=model_container, **self._extra_kwargs))
        return self

    def kwargs(self) -> Mapping[str, Any]:
        if self._final_kwargs is None:
            msg = "`load` must be called prior to calling `kwargs`."
            raise ValueError(msg)
        return self._final_kwargs


def dynamic_token_batching(
    batch: pyarrow.RecordBatch,
    *,
    token_budget: int,
    column_name: str,
    tokenizer: tokenization_utils_base.PreTrainedTokenizerBase,
    max_input_size: int,
) -> Iterator[pyarrow.RecordBatch]:
    """Split the batch into a set of microbatches.

    The logic for the split is as follows:
    1. split the batch into non-null batch and null batch
    2. if non-null batch has rows:
        a. tokenize the input column in the non-null batch
        b. the data should be sorted according to token count of the tokenized input column
        c. calculate the number of rows that can fit within our token budget w/ padding
        d. build the microbatch and yield it while maintaining other data in those rows
    3. if null batch has rows, yield the null batch as the last microbatch
    """

    batch, null_batch = _split_batch_by_null_value(batch, column_name)

    if batch.num_rows > 0:
        col = batch.column(column_name)
        encoded_col = tokenizer(col.to_pylist(), truncation=True, max_length=max_input_size)

        input_ids = numpy.array([numpy.array(seq) for seq in encoded_col["input_ids"]], dtype=object)
        token_type_ids = numpy.array([numpy.array(seq) for seq in encoded_col["token_type_ids"]], dtype=object)
        attention_mask = numpy.array([numpy.array(seq) for seq in encoded_col["attention_mask"]], dtype=object)

        lengths = numpy.fromiter(map(len, input_ids), dtype=int)
        sorted_indices = numpy.argsort(lengths)[::-1]

        index = 0
        while index < batch.num_rows:
            max_token_count = lengths[sorted_indices[index]]

            #  Calculate the number of rows we can support according to our token budget (assumes padding)
            #  This is simply `budget // max_token_count` since our data is sorted by length desc.
            #  NOTE: we ensure that we always take one item in the case that our token budget is smaller than
            #  a single row.
            num_rows = max((token_budget // max_token_count), 1)
            end = min(index + num_rows, batch.num_rows)
            microbatch_indices = sorted_indices[index:end]

            input_ids_microbatch = _pad_batch_to_length(
                input_ids.take(microbatch_indices, axis=0), max_token_count, tokenizer.pad_token_type_id
            )
            token_type_ids_microbatch = _pad_batch_to_length(
                token_type_ids.take(microbatch_indices, axis=0), max_token_count, tokenizer.pad_token_type_id
            )
            attention_mask_microbatch = _pad_batch_to_length(
                attention_mask.take(microbatch_indices, axis=0), max_token_count, tokenizer.pad_token_type_id
            )

            microbatch = batch.take(microbatch_indices)

            new_columns = [
                pyarrow.array(input_ids_microbatch),
                pyarrow.array(token_type_ids_microbatch),
                pyarrow.array(attention_mask_microbatch),
            ]
            new_schemas = [
                pyarrow.field(_INPUT_IDS_COL_NAME, pyarrow.list_(pyarrow.int64())),
                pyarrow.field(_TOKEN_TYPE_IDS_COL_NAME, pyarrow.list_(pyarrow.int64())),
                pyarrow.field(_ATTENTION_MASK_COL_NAME, pyarrow.list_(pyarrow.int64())),
            ]
            microbatch = pyarrow.RecordBatch.from_arrays(
                [*microbatch.columns, *new_columns],
                schema=pyarrow.schema(list(microbatch.schema) + new_schemas),
            )
            yield microbatch
            index = end

    if null_batch.num_rows > 0:
        yield null_batch


def _split_batch_by_null_value(
    batch: pyarrow.RecordBatch, column_name: str
) -> Tuple[pyarrow.RecordBatch, pyarrow.RecordBatch]:
    column = batch.column(column_name)
    null_filter = pc.is_null(column)
    non_null_filter = pc.invert(null_filter)

    null_batch = batch.filter(null_filter)
    non_null_batch = batch.filter(non_null_filter)

    return non_null_batch, null_batch


def _pad_batch_to_length(
    batch: np_typing.ArrayLike, max_length: int, padding_value: int, padding_side: str = "right"
) -> np_typing.ArrayLike:
    return [_pad_tokens_to_length(tokens, max_length, padding_value, padding_side) for tokens in batch]


def _pad_tokens_to_length(
    tokens: np_typing.ArrayLike, max_length: int, padding_value: int, padding_side: str = "right"
) -> np_typing.ArrayLike:
    if padding_side == "right":
        pad_width = (0, max_length - len(tokens))
    else:
        pad_width = (max_length - len(tokens), 0)

    return numpy.pad(tokens, pad_width, mode="constant", constant_values=padding_value)


@dataclass
class DynamicBatchingFuncConfig:
    _model_filename: str
    _token_budget: int
    _column_name: str

    # Note: _final_kwargs is a late init object.
    _final_kwargs: Optional[Mapping[str, Any]] = None

    @classmethod
    def create(cls, token_budget: int, column_name: str, model_filename: str) -> DynamicBatchingFuncConfig:
        return cls(_token_budget=token_budget, _column_name=column_name, _model_filename=model_filename)

    def load(self):
        model_container = ModelContainer.load(self._model_filename, device_name="cpu")
        self._final_kwargs = types.MappingProxyType(
            {
                "token_budget": self._token_budget,
                "column_name": self._column_name,
                "tokenizer": model_container.tokenizer,
                "max_input_size": model_container.config.max_input_size,
            }
        )
        return self

    def kwargs(self) -> Mapping[str, Any]:
        if self._final_kwargs is None:
            msg = "`load` must be called prior to calling `kwargs`."
            raise ValueError(msg)
        return self._final_kwargs



from __future__ import annotations

import threading
from typing import Iterator
from typing import List
from typing import Sequence
from typing import Tuple

import pyarrow

from tecton_core.embeddings import execution_utils
from tecton_core.vendor import queue


QUEUE_TIMEOUT_SECS = 5

# NOTE: we set a max record batch size when using multiple model inference workers
# to better parallelize across our workers. The default max record batch size we get
# from the DuckDB SQL output is 1_000_000, which is quite large for embeddings inference
# since embedding 1M strings could take tens of minutes on longer strings.
#
# WARN: Setting this lower is not necessarily a 'free' speedup, since smaller record batches
# on high variance token lengths gives us less opportunity to optimize the batch size.
MULTITHREADED_MAX_RECORD_BATCH_SIZE = 250_000


def data_coordinator_thread_func(
    data_iter: pyarrow.RecordBatchReader,
    output_queue: queue.ClosableQueue,
    max_record_batch_size: int = MULTITHREADED_MAX_RECORD_BATCH_SIZE,
) -> None:
    for data in data_iter:
        if max_record_batch_size is None or len(data) < max_record_batch_size:
            output_queue.put(data)
        else:
            for offset in range(0, len(data), max_record_batch_size):
                length = min(len(data) - offset, max_record_batch_size)
                output_queue.put(data.slice(offset, length))

    output_queue.close()


def iterator_over_queue(q: queue.ClosableQueue) -> Iterator:
    while True:
        try:
            yield q.get(timeout=QUEUE_TIMEOUT_SECS)
        except queue.Closed as e:
            # NOTE: we are done if our input is closed!
            break
        except queue.Empty:
            continue


def _inference_on_record_batch(
    batch: pyarrow.RecordBatch,
    preprocess_func: execution_utils.PreprocessorCallable,
    preprocess_kwargs: execution_utils.P.kwargs,
    inference_func: execution_utils.ModelCallable,
    inference_kwargs: execution_utils.P.kwargs,
) -> Tuple[List[pyarrow.RecordBatch], List[execution_utils.ExceptionWrapper]]:
    token_batches = execution_utils.data_preprocessor_one_step(batch, preprocess_func, preprocess_kwargs)
    inference_results: List[pyarrow.RecordBatch] = []
    exceptions: List[execution_utils.ExceptionWrapper] = []
    for token_batch in token_batches:
        result = execution_utils.model_inference_one_step(token_batch, inference_func, inference_kwargs)
        if isinstance(result, execution_utils.ExceptionWrapper):
            exceptions.append(result)
        else:
            inference_results.append(result)
    if len(inference_results) > 0:
        # The dynamic token batching in preprocessing will likely increase the number of batches relative to the inputs.
        # This is to consolidate the results and prevent the # of batches from increasing.
        inference_results = pyarrow.Table.from_batches(inference_results).combine_chunks().to_batches()
    return inference_results, exceptions


def threaded_inference_func(
    input_queue: queue.ClosableQueue,
    output_queue: queue.ClosableQueue,
    preprocess_info: Tuple[execution_utils.PreprocessorCallable, execution_utils.FuncConfig],
    inference_info: Tuple[execution_utils.ModelCallable, execution_utils.FuncConfig],
) -> None:
    preprocess_func, preprocess_config = preprocess_info
    inference_func, inference_func_config = inference_info
    preprocess_kwargs = preprocess_config.load().kwargs()
    inference_kwargs = inference_func_config.load().kwargs()

    for batch in iterator_over_queue(input_queue):
        inference_results, exceptions = _inference_on_record_batch(
            batch, preprocess_func, preprocess_kwargs, inference_func, inference_kwargs
        )
        for result in (*inference_results, *exceptions):
            output_queue.put(result)


def execute_multithreaded(
    data_source: pyarrow.RecordBatchReader,
    preprocess_info: Tuple[execution_utils.PreprocessorCallable, execution_utils.FuncConfig],
    inference_info: Tuple[execution_utils.ModelCallable, Sequence[execution_utils.FuncConfig]],
) -> Iterator[pyarrow.RecordBatch]:
    # NOTE: for now this assumes that all data fits in memory. We will re-evaluate that in the future.
    input_queue = queue.ClosableQueue()  # Queue between coordinator and preprocessors
    output_queue = queue.ClosableQueue()  # Queue between preprocessors and model workers

    inference_func, inference_func_configs = inference_info

    # Start data coordinator
    coordinator = threading.Thread(target=data_coordinator_thread_func, args=(data_source, input_queue))
    coordinator.start()

    # Start model workers
    model_workers = [
        threading.Thread(
            target=threaded_inference_func,
            args=(
                input_queue,
                output_queue,
                preprocess_info,
                (inference_func, func_config),
            ),
            daemon=True,
        )
        for func_config in inference_func_configs
    ]
    for worker in model_workers:
        worker.start()

    coordinator.join()

    for worker in model_workers:
        worker.join()

    num_items = output_queue.qsize()

    # NOTE: this line means that we cannot work with larger than memory data.
    # We will come back to this later.
    output_batches = [output_queue.get() for _ in range(num_items)]

    for output_item in output_batches:
        if isinstance(output_item, execution_utils.ExceptionWrapper):
            output_item.reraise()

    yield from output_batches


def execute_singlethreaded(
    data_source: pyarrow.RecordBatchReader,
    preprocess_info: Tuple[execution_utils.PreprocessorCallable, execution_utils.FuncConfig],
    inference_info: Tuple[execution_utils.ModelCallable, Sequence[execution_utils.FuncConfig]],
) -> Iterator[pyarrow.RecordBatch]:
    preprocess_func, preprocess_config = preprocess_info
    preprocess_kwargs = preprocess_config.load().kwargs()

    inference_func, inference_config = inference_info
    if len(inference_config) > 1:
        msg = "Single threaded execution only supports one inference configuration"
        raise ValueError(msg)
    inference_kwargs = inference_config[0].load().kwargs()

    for batch in data_source:
        results, exceptions = _inference_on_record_batch(
            batch, preprocess_func, preprocess_kwargs, inference_func, inference_kwargs
        )
        if len(exceptions) > 0:
            exceptions[0].reraise()
        for result in results:
            yield result



import logging
import threading
from typing import Optional

from google.protobuf.empty_pb2 import Empty

import tecton
from tecton._internals.metadata_service_impl.auth_lib import InternalAuthProvider
from tecton._internals.metadata_service_impl.request_lib import InternalRequestProvider
from tecton._internals.metadata_service_impl.service_modules import GRPC_SERVICE_MODULES
from tecton_core import conf
from tecton_core.metadata_service_impl.base_stub import BaseStub
from tecton_core.metadata_service_impl.http_client import PureHTTPStub


_lock = threading.Lock()
_stub_instance: Optional[BaseStub] = None

logger = logging.getLogger(__name__)


def instance() -> BaseStub:
    if getattr(tecton, "__initializing__"):
        msg = "Tried to initialize MDS during module import"
        raise Exception(msg)
    global _stub_instance
    with _lock:
        if not _stub_instance:
            stub = PureHTTPStub(InternalRequestProvider(InternalAuthProvider()), GRPC_SERVICE_MODULES)
            conf._init_metadata_server_config(stub.GetConfigs(Empty()))

            # Only set the global at the end. Otherwise if GetConfigs() fails, we won't try to initialize the configs
            # again if the user retries
            _stub_instance = stub
        return _stub_instance


def close_instance():
    global _stub_instance
    with _lock:
        if _stub_instance:
            _stub_instance.close()
            _stub_instance = None



import contextlib
import logging
from abc import ABC
from abc import abstractmethod
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import Union

import attrs
import pandas
import pyarrow

from tecton_core.offline_store import OfflineStoreOptionsProvider
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.schema import Schema
from tecton_core.secrets import SecretResolver


logger = logging.getLogger(__name__)


@attrs.define
class QueryTreeCompute(ABC):
    """
    Base class for compute (e.g. DWH compute or Python compute) which can be
    used for different stages of executing the query tree.
    """


@attrs.define
class ComputeMonitor:
    log_progress: Callable[[float], None] = lambda _: _  # noqa: E731
    set_query: Callable[[str], None] = lambda _: _  # noqa: E731


@attrs.define
class SQLCompute(QueryTreeCompute, contextlib.AbstractContextManager):
    """
    Base class for compute backed by a SQL engine (e.g. Snowflake and DuckDB).
    """

    @staticmethod
    def for_dialect(
        dialect: Dialect,
        qt_root: Optional[NodeRef] = None,
        secret_resolver: Optional[SecretResolver] = None,
        offline_store_options: Iterable[OfflineStoreOptionsProvider] = (),
    ) -> "SQLCompute":
        # Conditional imports are used so that optional dependencies such as the Snowflake connector are only imported
        # if they're needed for a query
        if dialect == Dialect.SNOWFLAKE:
            from tecton_core.query.snowflake.compute import SnowflakeCompute
            from tecton_core.query.snowflake.compute import create_snowflake_connection

            if SnowflakeCompute.is_context_initialized():
                return SnowflakeCompute.from_context()
            return SnowflakeCompute.for_connection(create_snowflake_connection(qt_root, secret_resolver))
        if dialect == Dialect.DUCKDB:
            from tecton_core.query.duckdb.compute import DuckDBCompute

            return DuckDBCompute.from_context(offline_store_options=offline_store_options)
        if dialect == Dialect.PANDAS:
            from tecton_core.query.pandas.compute import PandasCompute

            return PandasCompute.from_context()
        if dialect == Dialect.BIGQUERY:
            from tecton_core.query.bigquery.compute import BigqueryCompute

            return BigqueryCompute()

    @abstractmethod
    def get_dialect(self) -> Dialect:
        pass

    @abstractmethod
    def run_sql(
        self,
        sql_string: str,
        return_dataframe: bool = False,
        expected_output_schema: Optional[Schema] = None,
        monitor: Optional[ComputeMonitor] = None,
    ) -> Optional[pyarrow.RecordBatchReader]:
        pass

    @abstractmethod
    def register_temp_table(
        self, table_name: str, table_or_reader: Union[pyarrow.Table, pyarrow.RecordBatchReader]
    ) -> None:
        pass

    # TODO(danny): remove this once we convert connectors to return arrow tables instead of pandas dataframes
    @abstractmethod
    def register_temp_table_from_pandas(self, table_name: str, pandas_df: pandas.DataFrame) -> None:
        pass

    @abstractmethod
    def register_temp_table_from_data_source(
        self,
        table_name: str,
        ds: DataSourceScanNode,
        secret_resolver: Optional[SecretResolver] = None,
        monitor: Optional[ComputeMonitor] = None,
    ) -> None:
        pass

    @abstractmethod
    def load_table(self, table_name: str, expected_output_schema: Optional[Schema] = None) -> pyarrow.RecordBatchReader:
        pass

    def cleanup_temp_tables(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup_temp_tables()


@attrs.define
class ModelInferenceCompute(QueryTreeCompute, ABC):
    """
    Base class for compute that executes model inference (e.g. Torch).
    """

    @staticmethod
    def for_dialect(dialect: Dialect) -> "ModelInferenceCompute":
        if dialect == Dialect.TORCH:
            from tecton_core.embeddings.compute import TorchCompute

            return TorchCompute.from_context()

    @abstractmethod
    def run_inference(
        self, qt_node: NodeRef, input_data: Union[pyarrow.Table, pyarrow.RecordBatchReader]
    ) -> pyarrow.RecordBatchReader:
        pass



import json
import logging
import math
import shutil
import sys
import tempfile
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from dataclasses import dataclass
from datetime import datetime
from datetime import timezone
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple

import click
from google.protobuf import timestamp_pb2
from tqdm import tqdm

from tecton import version
from tecton._internals import metadata_service
from tecton.cli import printer
from tecton.cli.cli_utils import display_table
from tecton.cli.command import TectonGroup
from tecton.cli.environment_utils import EnvironmentDependencies
from tecton.cli.environment_utils import create_requirements_from_lock_file
from tecton.cli.environment_utils import download_dependencies
from tecton.cli.environment_utils import is_valid_environment_name
from tecton.cli.environment_utils import resolve_dependencies
from tecton_core import http
from tecton_core import id_helper
from tecton_proto.common.container_image__client_pb2 import ContainerImage
from tecton_proto.data.remote_compute_environment__client_pb2 import ObjectStoreUploadPart
from tecton_proto.data.remote_compute_environment__client_pb2 import RemoteComputeType
from tecton_proto.data.remote_compute_environment__client_pb2 import RemoteEnvironmentStatus
from tecton_proto.data.remote_compute_environment__client_pb2 import RemoteEnvironmentUploadInfo
from tecton_proto.data.remote_compute_environment__client_pb2 import S3UploadInfo
from tecton_proto.data.remote_compute_environment__client_pb2 import S3UploadPart
from tecton_proto.remoteenvironmentservice.remote_environment_service__client_pb2 import CompletePackagesUploadRequest
from tecton_proto.remoteenvironmentservice.remote_environment_service__client_pb2 import CreateRemoteEnvironmentRequest
from tecton_proto.remoteenvironmentservice.remote_environment_service__client_pb2 import DeleteRemoteEnvironmentsRequest
from tecton_proto.remoteenvironmentservice.remote_environment_service__client_pb2 import GetPackagesUploadUrlRequest
from tecton_proto.remoteenvironmentservice.remote_environment_service__client_pb2 import ListRemoteEnvironmentsRequest
from tecton_proto.remoteenvironmentservice.remote_environment_service__client_pb2 import StartPackagesUploadRequest


# libraries may have Requires: Python >=3.8.1
DEFAULT_PYTHON_VERSION = "3.8.17"
RESOLVED_REQUIREMENTS_FILENAME = "resolved_requirements.txt"
ERROR_MESSAGE_PREFIX = "⛔ ERROR: "
DEPENDENCY_RESOLUTION_TIMEOUT_SECONDS = 600
TECTON_RIFT_MATERIALIZATION_RUNTIME_PACKAGE = "tecton"
TECTON_TRANSFORM_RUNTIME_PACKAGE = "tecton-runtime"
DEFAULT_ARCHITECTURE = "x86_64"
CHECK_MARK = "✅"
ERROR_SIGN = "⛔"

# boto3 defaults to 8MB for multi-part uploads using upload_file.
DEFAULT_UPLOAD_PART_SIZE_MB = 16

# 5 was arbitrarily selected. We want to be conservative as this will run in customer's environments
DEFAULT_MAX_WORKERS_THREADS = 5

# The maximum size of all dependencies allowed for upload
MAX_ALLOWED_DEPENDENCIES_SIZE_GB = 10

MEGABYTE = 1024 * 1024
GIGABYTE = 1024 * MEGABYTE

logger = logging.getLogger(__name__)


@dataclass
class UploadPart:
    """
    Represents an individual part of a file that needs to be uploaded in chunks or parts.
    :param part_number (int): The 1-indexed number of the part to be uploaded.
    :param offset (int): The starting byte offset of this part in the file.
    :param part_size (int): The size of this part in bytes.
    """

    part_number: int
    offset: int
    part_size: int


@dataclass
class EnvironmentIdentifier:
    id: Optional[str]
    name: Optional[str]

    def __post_init__(self):
        if not self.id and not self.name:
            printer.safe_print(
                f"{ERROR_MESSAGE_PREFIX} At least one of `environment-id` or `name` must be provided", file=sys.stderr
            )
            sys.exit(1)

    def __str__(self):
        if self.id:
            return f"id: {self.id}"
        elif self.name:
            return f"name: {self.name}"
        else:
            return "No name or id set"

    def __eq__(self, identifier):
        if isinstance(identifier, EnvironmentIdentifier):
            if self.id:
                return self.id == identifier.id
            elif self.name:
                return self.name == identifier.name
        return False


@click.command("environment", cls=TectonGroup)
def environment():
    """Manage Environments for ODFV execution.

    Use `tecton --verbose environment` to print additional debug information with environment commands.
    """


@environment.command("list")
def list():
    """List all available Python Environments"""
    remote_environments = _list_environments()
    _display_environments(remote_environments)


@environment.command("get")
@click.option("--environment-id", help="Environment Id", required=False, type=str)
@click.option("--name", help="Environment Name", required=False, type=str)
def get(environment_id: Optional[str] = None, name: Optional[str] = None):
    """Get Python Environment(s) matching a name or an ID"""
    environment_identifier = EnvironmentIdentifier(id=environment_id, name=name)
    remote_environments = _list_environments(environment_identifier=environment_identifier)
    if len(remote_environments) < 1:
        error_message = f"⛔ Could not find a match for environment with {environment_identifier.__str__()}!"
        printer.safe_print(error_message, file=sys.stderr)
        sys.exit(1)
    _display_environments(remote_environments)


@environment.command("resolve-dependencies")
@click.option("-r", "--requirements", help="Path to a requirements file", required=True, type=click.Path(exists=True))
@click.option(
    "-o",
    "--output-file",
    help="Output file to write resolved and fully pinned requirements to. If not specified, the pinned requirements will be printed to stdout",
    required=False,
    type=click.Path(exists=False),
)
@click.option(
    "-p",
    "--python-version",
    help=f"Python Version for the environment, defaults to {DEFAULT_PYTHON_VERSION}",
    required=False,
)
def resolve_requirements(
    requirements: str,
    output_file: Optional[str] = None,
    python_version: Optional[str] = None,
):
    """Resolve dependencies and return a fully resolved set of requirements for a given requirements.txt"""
    _python_version = python_version or DEFAULT_PYTHON_VERSION

    with tempfile.TemporaryDirectory() as tmpdir:
        try:
            resolved_requirements_path, lock_output_path = _run_dependency_resolution(
                requirements_path=Path(requirements),
                resolved_requirements_directory=Path(tmpdir),
                python_version=_python_version,
            )
        except ValueError as e:
            printer.safe_print(f"{ERROR_MESSAGE_PREFIX} {e}", file=sys.stderr)
            sys.exit(1)

        _validate_and_get_dependencies(lock_output_path)
        printer.safe_print("✅ Successfully resolved dependencies")

        if output_file is not None:
            output_path = Path(output_file)
            resolved_requirements_str = resolved_requirements_path.read_bytes()
            output_path.write_bytes(resolved_requirements_str)
        else:
            printer.safe_print("\n🎉 Fully Resolved Requirements: \n")
            _display_requirements(requirements_path=resolved_requirements_path)


@environment.command("create")
@click.option("-n", "--name", help="Name of the environment", required=True, type=str)
@click.option(
    "-r",
    "--requirements",
    help="Path to the requirements.txt file containing all dependencies for the environment",
    required=True,
    type=click.Path(exists=True),
)
@click.option("-d", "--description", help="A description for the environment", required=False, type=str)
@click.option(
    "-p",
    "--python-version",
    help=f"Python Version for the environment, defaults to {DEFAULT_PYTHON_VERSION}",
    required=False,
)
def create(
    name: str,
    requirements: str,
    description: Optional[str] = None,
    python_version: Optional[str] = None,
):
    """Create a custom Python Environment for ODFVs."""
    return _create(name, requirements, description, python_version)


def _create(
    name: str,
    requirements: str,
    description: Optional[str] = None,
    python_version: Optional[str] = None,
):
    """Create a custom Python Environment for ODFVs"""
    _python_version = python_version or DEFAULT_PYTHON_VERSION
    if not is_valid_environment_name(name):
        printer.safe_print(
            f"{ERROR_MESSAGE_PREFIX} Invalid name. Custom environment names can only contain letters, numbers, hyphens, and underscores",
            file=sys.stderr,
        )
        sys.exit(1)

    environment_names = [e.name for e in _list_environments()]
    if name in environment_names:
        printer.safe_print(
            f"{ERROR_MESSAGE_PREFIX} An environment with the name `{name}` already exists in Tecton!",
            file=sys.stderr,
        )
        sys.exit(1)

    resp = _create_environment_with_requirements(
        name,
        description,
        Path(requirements),
        _python_version,
    )
    if resp:
        _display_environments([resp.remote_environment])
        printer.safe_print(
            f"\n🎉 Successfully created environment {name} with Status=PENDING. Please run `tecton environment get --name <environment-name>` to monitor the status of the environment"
        )


@environment.command("describe")
@click.option("-i", "--environment-id", help="Environment ID", required=False, type=str)
@click.option("-n", "--name", help="Environment Name", required=False, type=str)
def describe(environment_id: Optional[str] = None, name: Optional[str] = None):
    """
    Print additional information about an environment
    """
    environment_identifier = EnvironmentIdentifier(id=environment_id, name=name)
    environments = _list_environments(environment_identifier)
    if not environments:
        error_message = f"⛔ Could not find a match for environment with {environment_identifier.__str__()}!"
        printer.safe_print(error_message, file=sys.stderr)
        sys.exit(1)
    if len(environments) > 1:
        message = (
            f"Could not find environment with {environment_identifier.__str__()}. Did you mean one of the following?"
        )
        printer.safe_print(f"⚠️ {message}")
        _display_environments(environments)
    else:
        environment_match = environments[0]
        printer.safe_print("\n💡 Environment Details: \n")
        _display_environments(environments)
        printer.safe_print("\n💡 Input Requirements: \n")
        _display_requirements(requirements_str=environment_match.requirements)
        printer.safe_print("\n✅ Fully Resolved Requirements: \n")
        _display_requirements(requirements_str=environment_match.resolved_requirements)


@environment.command("delete")
@click.option("-i", "--environment-id", help="Environment ID", required=False, type=str)
@click.option("-n", "--name", help="Environment Name", required=False, type=str)
def delete(environment_id: Optional[str] = None, name: Optional[str] = None):
    """Delete an existing custom Python Environment by name or ID"""
    return _delete(environment_id=environment_id, name=name)


def _delete(environment_id: Optional[str] = None, name: Optional[str] = None, skip_confirmation: bool = False):
    """Delete an existing custom Python Environment by name or ID"""
    environment_identifier = EnvironmentIdentifier(id=environment_id, name=name)
    environments = _list_environments(environment_identifier=environment_identifier)
    if not environments:
        printer.safe_print(
            f"⛔ No matching environment found for: {environment_identifier.__str__()}. Please verify available environments using the `tecton environment list` command",
            file=sys.stderr,
        )
        sys.exit(1)
    result_identifier = None
    environment_to_delete = None
    if len(environments) == 1:
        environment_to_delete = environments[0]
        result_identifier = EnvironmentIdentifier(id=environment_to_delete.id, name=environment_to_delete.name)

    if len(environments) > 1 or not environment_identifier.__eq__(identifier=result_identifier):
        printer.safe_print(
            f"⚠️ No matching environment found for: {environment_identifier.__str__()}. Did you mean one of the following environment(s)? \n\n",
            file=sys.stderr,
        )
        _display_environments(environments)
    else:
        confirmation_text = f"⚠️  Are you sure you want to delete environment {environment_to_delete.name}? (y/n) :"
        confirmation = "y" if skip_confirmation else input(confirmation_text).lower().strip()
        if confirmation == "y":
            try:
                _delete_environment(environment_id=environment_to_delete.id)
                printer.safe_print(f"✅ Marked environment '{environment_identifier.__str__()}' for deletion.")
            except Exception as e:
                printer.safe_print(f"⛔ Failed to delete environment. error = {str(e)}, type= {type(e).__name__}")
        else:
            printer.safe_print(f"Cancelled deletion for environment: {environment_identifier.__str__()}")


def _display_environments(environments: list):
    headings = [
        "Id",
        "Name",
        "Description",
        "Type",
        "Materialization Version",
        "Tecton Runtime Version",
        "Created At",
        "Status",
        "Status Details",
    ]
    display_table(
        headings,
        [
            (
                i.id,
                i.name,
                i.description,
                RemoteComputeType.Name(i.type).split("_")[-1],
                i.rift_batch_job_environment.tecton_materialization_runtime_version or "N/A",
                i.realtime_job_environment.tecton_transform_runtime_version or "N/A",
                _timestamp_to_string(i.created_at),
                RemoteEnvironmentStatus.Name(i.status).split("_")[-1],
                i.status_details or "OK",
            )
            for i in environments
        ],
    )


def _display_dependent_feature_services(dependent_feature_services: list):
    headings = ["Workspace Name", "Feature Service Name"]
    display_table(
        headings,
        [
            (
                i.workspace_name,
                i.feature_service_name,
            )
            for i in dependent_feature_services
        ],
    )


def _create_environment_with_image(name: str, description: str, image_uri):
    try:
        req = CreateRemoteEnvironmentRequest()
        req.name = name
        req.description = description

        image_info = ContainerImage()
        image_info.image_uri = image_uri

        req.image_info.CopyFrom(image_info)

        return metadata_service.instance().CreateRemoteEnvironment(req)
    except PermissionError as _:
        printer.safe_print(
            "The user is not authorized to create environment(s) in Tecton. Please reach out to your Admin to complete this "
            "action",
            file=sys.stderr,
        )
        sys.exit(1)
    except Exception as e:
        printer.safe_print(f"Failed to create environment: {e}", file=sys.stderr)
        sys.exit(1)


def _create_environment_with_requirements(
    name: str,
    description: str,
    requirements_path: Path,
    python_version: str,
):
    """Create a custom environment by resolving dependencies, downloading wheels and updating MDS
    Parameters:
        name(str): Name of the custom environment
        description(str): Description of the custom environment
        requirements_path(str): Path to the `requirements.txt` file
        python_version(str): The Python version to resolve the dependencies for
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        try:
            resolved_requirements_path, lock_output_path = _run_dependency_resolution(
                requirements_path=requirements_path,
                resolved_requirements_directory=Path(tmpdir),
                python_version=python_version,
            )
        except ValueError as e:
            printer.safe_print(f"{ERROR_MESSAGE_PREFIX} {e}", file=sys.stderr)
            sys.exit(1)

        dependencies = _validate_and_get_dependencies(lock_output_path)

        printer.safe_print("✅ Successfully resolved dependencies")
        tecton_runtime_version = dependencies.get_version(dependency_name=TECTON_TRANSFORM_RUNTIME_PACKAGE)
        tecton_rift_version = dependencies.get_version(dependency_name=TECTON_RIFT_MATERIALIZATION_RUNTIME_PACKAGE)

        printer.safe_print(
            f"\n💡 Creating environment '{name}' for job types:\n"
            f"{CHECK_MARK if tecton_runtime_version else ERROR_SIGN} On Demand\n"
            f"{CHECK_MARK if tecton_rift_version else ERROR_SIGN} Rift Batch\n"
            f"{CHECK_MARK if tecton_rift_version and tecton_runtime_version else ERROR_SIGN} Rift Stream Ingest\n"
        )

        url_requirements_path = Path(tmpdir) / "url_requirements.txt"

        # doing this to use the url because some wheels maybe coming from different sources
        create_requirements_from_lock_file(lock_output_path, url_requirements_path)

        download_wheels_dir = Path(tmpdir) / "wheels"
        download_wheels_dir.mkdir()
        printer.safe_print("\n⏳ Downloading wheels. This may take a few seconds.....\n")
        download_dependencies(
            requirements_path=url_requirements_path,
            target_directory=download_wheels_dir,
            python_version=python_version,
        )
        printer.safe_print("\n✅ Successfully downloaded dependencies")

        directory_size = _get_directory_size(download_wheels_dir)
        if directory_size > (MAX_ALLOWED_DEPENDENCIES_SIZE_GB * GIGABYTE):
            printer.safe_print(
                f"{ERROR_MESSAGE_PREFIX} The total size of the downloaded dependencies exceeds the max allowed limit of {MAX_ALLOWED_DEPENDENCIES_SIZE_GB}GB. Please reduce the total number / size of dependencies and try again!",
                file=sys.stderr,
            )
            sys.exit(1)

        printer.safe_print("\n⏳ Uploading compressed wheels in parts to S3. This may take a few seconds.....")
        environment_id = id_helper.IdHelper.generate_string_id()
        try:
            location = _upload_dependencies(source_path=download_wheels_dir, environment_id=environment_id)
        except ValueError as e:
            printer.safe_print(f"{ERROR_MESSAGE_PREFIX} Unable to upload dependencies - {e}", file=sys.stderr)
            sys.exit(1)

        req = CreateRemoteEnvironmentRequest(
            name=name,
            id=environment_id,
            description=description,
            python_version=python_version,
            s3_wheels_location=location,
            requirements=requirements_path.read_text(),
            resolved_requirements=resolved_requirements_path.read_text(),
            transform_runtime_version=tecton_runtime_version,
            rift_materialization_runtime_version=tecton_rift_version,
            sdk_version=version.get_semantic_version(),
        )
        return metadata_service.instance().CreateRemoteEnvironment(req)


def _validate_and_get_dependencies(lock_file_path: Path) -> EnvironmentDependencies:
    with open(lock_file_path, "r") as lock_file:
        dependencies_data = json.load(lock_file)
    dependencies = EnvironmentDependencies(dependencies_data)
    tecton_runtime_version = dependencies.get_version(dependency_name=TECTON_TRANSFORM_RUNTIME_PACKAGE)
    tecton_rift_version = dependencies.get_version(dependency_name=TECTON_RIFT_MATERIALIZATION_RUNTIME_PACKAGE)

    if tecton_rift_version is not None:
        rift_extras = dependencies.get_dependency_extras(TECTON_RIFT_MATERIALIZATION_RUNTIME_PACKAGE)
        if "rift-materialization" not in rift_extras:
            printer.safe_print(
                f"{ERROR_MESSAGE_PREFIX} `rift-materialization` extra must be included for package `tecton` to support Rift materialization (tecton[rift-materialization]==x.x.x)",
                file=sys.stderr,
            )
            sys.exit(1)

    if not tecton_runtime_version and not tecton_rift_version:
        # TODO (vitaly): update the doc link for more information on custom Rift materialization envs
        printer.safe_print(
            f"{ERROR_MESSAGE_PREFIX} Please include at least one of the supporting Tecton libraries:\n"
            f"💡  `tecton-runtime` package (https://pypi.org/project/tecton-runtime) to support On Demand environments\n"
            f"💡  `tecton[rift-materialization]` package (https://pypi.org/project/tecton) to support Rift materialization environments",
            "\nFor more information, please see: https://docs.tecton.ai/docs/defining-features/feature-views/on-demand-feature-view/on-demand-feature-view-environments/custom-on-demand-feature-view-environments",
            file=sys.stderr,
        )
        sys.exit(1)
    return dependencies


def _run_dependency_resolution(
    requirements_path: Path, resolved_requirements_directory: Path, python_version: str
) -> Tuple[Path, Path]:
    printer.safe_print(
        f"\n⏳ Resolving dependencies for Python {python_version} and architecture {DEFAULT_ARCHITECTURE}. This may take a few seconds....."
    )
    lock_output_path = resolved_requirements_directory / "lock.json"
    resolved_requirements_path = resolved_requirements_directory / RESOLVED_REQUIREMENTS_FILENAME
    resolve_dependencies(
        python_version=python_version,
        requirements_path=requirements_path,
        lock_output_path=lock_output_path,
        resolved_requirements_path=resolved_requirements_path,
        timeout_seconds=DEPENDENCY_RESOLUTION_TIMEOUT_SECONDS,
    )
    return resolved_requirements_path, lock_output_path


def _delete_environment(environment_id: str):
    try:
        req = DeleteRemoteEnvironmentsRequest()
        req.ids.append(environment_id)
        return metadata_service.instance().DeleteRemoteEnvironments(req)
    except PermissionError as _:
        printer.safe_print(
            "⛔ The user is not authorized to perform environment deletion. Please reach out to your Admin to complete this action.",
            file=sys.stderr,
        )
        sys.exit(1)
    except Exception as e:
        printer.safe_print(f"⛔ Failed to delete environment: {e}", file=sys.stderr)
        sys.exit(1)


def _list_environments(environment_identifier: Optional[EnvironmentIdentifier] = None):
    try:
        req = ListRemoteEnvironmentsRequest()
        response = metadata_service.instance().ListRemoteEnvironments(req)
        if not environment_identifier:
            return response.remote_environments
        if environment_identifier.id:
            environments = [env for env in response.remote_environments if environment_identifier.id in env.id]
        else:
            # Look for an exact match. If there are no exact matches, we will return all substring matches
            environments = [env for env in response.remote_environments if environment_identifier.name == env.name]
            if not environments:
                environments = [env for env in response.remote_environments if environment_identifier.name in env.name]

        return environments

    except Exception as e:
        printer.safe_print(f"Failed to fetch environments: {e}", file=sys.stderr)
        sys.exit(1)


def _timestamp_to_string(value: timestamp_pb2.Timestamp) -> str:
    t = datetime.fromtimestamp(value.ToSeconds())
    return t.astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M:%S %Z")


def _upload_dependencies(source_path: Path, environment_id: str) -> str:
    """Upload dependencies from the specified source path to S3.

    Args:
        source_path (str): The path to the dependencies to upload.
        environment_id (str): The ID of the environment.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        output_zip_file = Path(tmpdir) / "wheels.zip"
        logger.debug(f"Zipping dependencies at {output_zip_file}")

        shutil.make_archive(str(output_zip_file.with_suffix("")), "zip", str(source_path))
        file_size = output_zip_file.stat().st_size

        logger.debug("Initiating Multi-Part Upload")
        start_request = StartPackagesUploadRequest(environment_id=environment_id)
        start_response = metadata_service.instance().StartPackagesUpload(start_request)

        upload_id = start_response.upload_info.s3_upload_info.upload_id
        upload_parts = _upload_file_in_parts(
            file_size=file_size,
            upload_id=upload_id,
            environment_id=environment_id,
            output_zip_file=output_zip_file,
        )

        complete_request = CompletePackagesUploadRequest(
            upload_info=RemoteEnvironmentUploadInfo(
                environment_id=environment_id,
                s3_upload_info=S3UploadInfo(upload_id=upload_id, upload_parts=upload_parts),
            )
        )
        complete_response = metadata_service.instance().CompletePackagesUpload(complete_request)
        location = complete_response.storage_location
        printer.safe_print("✅ Successfully uploaded dependencies")
        return location


def _upload_file_in_parts(
    file_size: int, upload_id: str, environment_id: str, output_zip_file: Path
) -> List[S3UploadPart]:
    """Upload a file in parallel, dividing it into parts.

    Args:
        file_size (int): The size of the file in bytes.
        upload_id (str): A unique identifier for the file upload, returned by S3.
        environment_id (str): The ID of the environment.
        output_zip_file (str): The path to the file to upload.

    Returns:
        list: A list of upload part results.
    """
    # Calculate all parts for multi part upload
    part_data_list = get_upload_parts(file_size=file_size)
    with ThreadPoolExecutor(DEFAULT_MAX_WORKERS_THREADS) as executor:
        upload_futures = [
            executor.submit(
                _upload_part,
                upload_part=part_data,
                parent_upload_id=upload_id,
                environment_id=environment_id,
                dependency_file_path=output_zip_file,
            )
            for part_data in part_data_list
        ]
        with tqdm(total=len(part_data_list), desc="Upload progress", ncols=100) as pbar:
            for future in as_completed(upload_futures):
                # Increment the tqdm progress bar whenever a future is done
                if future.result():
                    pbar.update(1)

        return [future.result() for future in upload_futures]


def get_upload_parts(file_size: int) -> List[UploadPart]:
    """
    Calculate UploadPart for each part of a file to be uploaded, given total file size.
    It considers the DEFAULT_UPLOAD_PART_SIZE_MB as the maximum size of each part.

    Args:
        file_size (int): The total size of the file being uploaded in bytes.

    Returns:
        List[UploadPart]: An list of UploadPart representing all parts to be uploaded with its part number,
                    starting offset, and size in bytes.
    """
    total_parts = _calculate_part_count(file_size, DEFAULT_UPLOAD_PART_SIZE_MB)
    chunk_size = DEFAULT_UPLOAD_PART_SIZE_MB * MEGABYTE
    upload_parts = []
    for i in range(1, total_parts + 1):
        offset = chunk_size * (i - 1)
        bytes_remaining = file_size - offset
        # Adjust the size for the last part if the remaining bytes are less than the DEFAULT_UPLOAD_PART_SIZE_MB
        current_chunk_size = chunk_size if bytes_remaining > chunk_size else bytes_remaining
        upload_parts.append(UploadPart(part_number=i, offset=offset, part_size=current_chunk_size))
    return upload_parts


def _get_directory_size(directory: Path) -> int:
    """
    Compute the size of a directory in bytes.

    Args:
        directory (Path): The directory path for which to compute the size.

    Returns:
        int: The size of the directory in bytes.
    """
    return sum(f.stat().st_size for f in directory.rglob("*") if f.is_file())


def _calculate_part_count(file_size: int, part_size_mb: int) -> int:
    """Calculate the number of parts the file will be divided into for uploading.

    Args:
        file_size (int): The size of the file
        part_size_mb (int): The size of each part in megabytes.

    Returns:
        int: The total number of parts.
    """
    chunk_size = part_size_mb * 1024 * 1024
    return int(math.ceil(file_size / chunk_size))


def _upload_part(
    upload_part: UploadPart,
    parent_upload_id: str,
    environment_id: str,
    dependency_file_path: str,
):
    """Upload a part of a file.

    Args:
        upload_part (UploadPart): The part to upload.
        parent_upload_id (str): The ID of the parent upload.
        environment_id (str): The ID of the environment.
        dependency_file_path (str): The path to the file to upload.

    Returns:
        S3UploadPart: An object representing the uploaded part.
    """
    request = GetPackagesUploadUrlRequest(
        environment_id=environment_id,
        upload_part=ObjectStoreUploadPart(
            s3_upload_part=S3UploadPart(parent_upload_id=parent_upload_id, part_number=upload_part.part_number)
        ),
    )
    response = metadata_service.instance().GetPackagesUploadUrl(request)
    signed_url = response.upload_url

    with open(dependency_file_path, "rb") as fp:
        fp.seek(upload_part.offset)
        file_data = fp.read(upload_part.part_size)
        response = http.session().put(signed_url, data=file_data)
        if response.ok:
            e_tag = response.headers["ETag"]
            return S3UploadPart(part_number=upload_part.part_number, e_tag=e_tag, parent_upload_id=parent_upload_id)
        else:
            msg = f"Upload failed with status {response.status_code} and error {response.text}"
            raise ValueError(msg)


def _display_requirements(requirements_path: Optional[Path] = None, requirements_str: Optional[str] = None):
    """
    Display requirements from a requirements.txt file after removing hashes from each line, if exists.

    Args:
        requirements_path (Path): Path to the requirements file.
        requirements_str (str): Contents of a requirements.txt file
    """
    cleaned_lines = []
    if requirements_path is not None:
        with requirements_path.open("r") as file:
            cleaned_lines = [line.strip() for line in file.readlines()]
    elif requirements_str is not None:
        cleaned_lines = [line.strip() for line in requirements_str.split("\n")]

    # Skip hashes and empty lines.
    cleaned_lines = [line.rstrip("\\").strip() for line in cleaned_lines if line and not line.startswith("--hash")]
    printer.safe_print("\n".join(cleaned_lines))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/online_store/feature_value__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5tecton_proto/online_store/feature_value__client.proto\x12\x19tecton_proto.online_store\x1a\x1fgoogle/protobuf/timestamp.proto\"\xf0\x03\n\x0c\x46\x65\x61tureValue\x12H\n\x0b\x61rray_value\x18\x01 \x01(\x0b\x32%.tecton_proto.online_store.ArrayValueH\x00R\narrayValue\x12%\n\rfloat64_value\x18\x02 \x01(\x01H\x00R\x0c\x66loat64Value\x12%\n\rfloat32_value\x18\x03 \x01(\x02H\x00R\x0c\x66loat32Value\x12!\n\x0bint64_value\x18\x04 \x01(\x03H\x00R\nint64Value\x12\x1f\n\nbool_value\x18\x05 \x01(\x08H\x00R\tboolValue\x12#\n\x0cstring_value\x18\x06 \x01(\tH\x00R\x0bstringValue\x12\x45\n\nnull_value\x18\x07 \x01(\x0b\x32$.tecton_proto.online_store.NullValueH\x00R\tnullValue\x12K\n\x0cstruct_value\x18\x08 \x01(\x0b\x32&.tecton_proto.online_store.StructValueH\x00R\x0bstructValue\x12\x42\n\tmap_value\x18\t \x01(\x0b\x32#.tecton_proto.online_store.MapValueH\x00R\x08mapValueB\x07\n\x05value\"b\n\x10\x46\x65\x61tureValueList\x12N\n\x0e\x66\x65\x61ture_values\x18\x01 \x03(\x0b\x32\'.tecton_proto.online_store.FeatureValueR\rfeatureValues\"k\n\x15\x42\x61tchCompactedDataRow\x12R\n\x0e\x63ompacted_tile\x18\x01 \x01(\x0b\x32+.tecton_proto.online_store.FeatureValueListR\rcompactedTile\"o\n\x17\x42\x61tchCompactedDataRowV2\x12T\n\x0f\x63ompacted_tiles\x18\x01 \x03(\x0b\x32+.tecton_proto.online_store.FeatureValueListR\x0e\x63ompactedTiles\"\xaa\x01\n\rRedisTAFVData\x12\x1f\n\x0b\x61nchor_time\x18\x01 \x01(\x03R\nanchorTime\x12(\n\x10written_by_batch\x18\x02 \x01(\x08R\x0ewrittenByBatch\x12N\n\x0e\x66\x65\x61ture_values\x18\x03 \x03(\x0b\x32\'.tecton_proto.online_store.FeatureValueR\rfeatureValues\"\x0b\n\tNullValue\"\xd1\x03\n\nArrayValue\x12#\n\rstring_values\x18\x01 \x03(\tR\x0cstringValues\x12H\n\x0c\x61rray_values\x18\x02 \x03(\x0b\x32%.tecton_proto.online_store.ArrayValueR\x0b\x61rrayValues\x12K\n\rstruct_values\x18\x03 \x03(\x0b\x32&.tecton_proto.online_store.StructValueR\x0cstructValues\x12\x42\n\nmap_values\x18\x04 \x03(\x0b\x32#.tecton_proto.online_store.MapValueR\tmapValues\x12\x1f\n\x0b\x62ool_values\x18\x05 \x03(\x08R\nboolValues\x12%\n\x0cint64_values\x18\x06 \x03(\x03\x42\x02\x10\x01R\x0bint64Values\x12)\n\x0e\x66loat32_values\x18\x07 \x03(\x02\x42\x02\x10\x01R\rfloat32Values\x12)\n\x0e\x66loat64_values\x18\x08 \x03(\x01\x42\x02\x10\x01R\rfloat64Values\x12%\n\x0cnull_indices\x18\t \x03(\x05\x42\x02\x10\x01R\x0bnullIndices\"N\n\x0bStructValue\x12?\n\x06values\x18\x01 \x03(\x0b\x32\'.tecton_proto.online_store.FeatureValueR\x06values\"\x84\x01\n\x08MapValue\x12\x39\n\x04keys\x18\x01 \x01(\x0b\x32%.tecton_proto.online_store.ArrayValueR\x04keys\x12=\n\x06values\x18\x02 \x01(\x0b\x32%.tecton_proto.online_store.ArrayValueR\x06values\"\x8c\x02\n\x11\x43\x61\x63hedFeatureView\x12N\n\x0e\x66\x65\x61ture_values\x18\x01 \x03(\x0b\x32\'.tecton_proto.online_store.FeatureValueR\rfeatureValues\x12)\n\x10\x66\x65\x61ture_statuses\x18\x04 \x03(\x05R\x0f\x66\x65\x61tureStatuses\x12\x41\n\x0e\x65\x66\x66\x65\x63tive_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\reffectiveTime\x12\x39\n\nupdated_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tupdatedAtB\x1a\n\x16\x63om.tecton.onlinestoreP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.online_store.feature_value__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.tecton.onlinestoreP\001'
  _ARRAYVALUE.fields_by_name['int64_values']._options = None
  _ARRAYVALUE.fields_by_name['int64_values']._serialized_options = b'\020\001'
  _ARRAYVALUE.fields_by_name['float32_values']._options = None
  _ARRAYVALUE.fields_by_name['float32_values']._serialized_options = b'\020\001'
  _ARRAYVALUE.fields_by_name['float64_values']._options = None
  _ARRAYVALUE.fields_by_name['float64_values']._serialized_options = b'\020\001'
  _ARRAYVALUE.fields_by_name['null_indices']._options = None
  _ARRAYVALUE.fields_by_name['null_indices']._serialized_options = b'\020\001'
  _FEATUREVALUE._serialized_start=118
  _FEATUREVALUE._serialized_end=614
  _FEATUREVALUELIST._serialized_start=616
  _FEATUREVALUELIST._serialized_end=714
  _BATCHCOMPACTEDDATAROW._serialized_start=716
  _BATCHCOMPACTEDDATAROW._serialized_end=823
  _BATCHCOMPACTEDDATAROWV2._serialized_start=825
  _BATCHCOMPACTEDDATAROWV2._serialized_end=936
  _REDISTAFVDATA._serialized_start=939
  _REDISTAFVDATA._serialized_end=1109
  _NULLVALUE._serialized_start=1111
  _NULLVALUE._serialized_end=1122
  _ARRAYVALUE._serialized_start=1125
  _ARRAYVALUE._serialized_end=1590
  _STRUCTVALUE._serialized_start=1592
  _STRUCTVALUE._serialized_end=1670
  _MAPVALUE._serialized_start=1673
  _MAPVALUE._serialized_end=1805
  _CACHEDFEATUREVIEW._serialized_start=1808
  _CACHEDFEATUREVIEW._serialized_end=2076
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/diff_test__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import basic_info__client_pb2 as tecton__proto_dot_args_dot_basic__info____client__pb2
from tecton_proto.args import data_source__client_pb2 as tecton__proto_dot_args_dot_data__source____client__pb2
from tecton_proto.args import diff_options__client_pb2 as tecton__proto_dot_args_dot_diff__options____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)tecton_proto/args/diff_test__client.proto\x12\x11tecton_proto.args\x1a*tecton_proto/args/basic_info__client.proto\x1a+tecton_proto/args/data_source__client.proto\x1a,tecton_proto/args/diff_options__client.proto\x1a$tecton_proto/common/id__client.proto\"?\n\x0b\x44iffTestFoo\x12\x17\n\x07\x66ield_a\x18\x01 \x01(\tR\x06\x66ieldA\x12\x17\n\x07\x66ield_b\x18\x02 \x01(\tR\x06\x66ieldB\"\xd1\t\n\x0c\x44iffTestArgs\x12\x39\n\x0ctest_args_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\ntestArgsId\x12\x30\n\x04info\x18\x02 \x01(\x0b\x32\x1c.tecton_proto.args.BasicInfoR\x04info\x12\x42\n\tnew_field\x18\x03 \x01(\x0b\x32\x1e.tecton_proto.args.DiffTestFooB\x05\x92M\x02\x08\x03R\x08newField\x12\x42\n\told_field\x18\x04 \x01(\x0b\x32\x1e.tecton_proto.args.DiffTestFooB\x05\x92M\x02\x08\x04R\x08oldField\x12*\n\rpassive_field\x18\x05 \x01(\tB\x05\x92M\x02\x08\x05R\x0cpassiveField\x12\x45\n\x1brecreate_suppressable_field\x18\x06 \x01(\tB\x05\x92M\x02\x08\x06R\x19recreateSuppressableField\x12z\n\"recreate_suppressable_nested_field\x18\x08 \x01(\x0b\x32&.tecton_proto.args.DiffTestNestedInnerB\x05\x92M\x02\x08\x06R\x1frecreateSuppressableNestedField\x12q\n2recreate_suppressable_invalidate_checkpoints_field\x18\t \x01(\tB\x05\x92M\x02\x08\x07R.recreateSuppressableInvalidateCheckpointsField\x12\xa6\x01\n9recreate_suppressable_invalidate_checkpoints_nested_field\x18\n \x01(\x0b\x32&.tecton_proto.args.DiffTestNestedInnerB\x05\x92M\x02\x08\x07R4recreateSuppressableInvalidateCheckpointsNestedField\x12\x61\n*recreate_suppressable_restart_stream_field\x18\x0b \x01(\tB\x05\x92M\x02\x08\x08R&recreateSuppressableRestartStreamField\x12@\n\x1cunannotated_needing_recreate\x18\x07 \x01(\tR\x1aunannotatedNeedingRecreate\x12\x33\n\x07options\x18\x0c \x03(\x0b\x32\x19.tecton_proto.args.OptionR\x07options\x12\x61\n\x17renamed_primitive_field\x18\r \x01(\tB)\x92M\x1a*\x18\n\x16\x66ormer_primitive_field\x92M\t*\x07\x12\x05\x30.6.0R\x15renamedPrimitiveField\x12\x83\x01\n\x15renamed_message_field\x18\x0e \x01(\x0b\x32&.tecton_proto.args.DiffTestNestedInnerB\'\x92M\x18*\x16\n\x14\x66ormer_message_field\x92M\t*\x07\x12\x05\x30.6.0R\x13renamedMessageField\"\x98\x02\n\x13\x44iffTestNestedInner\x12*\n\rinplace_field\x18\x01 \x01(\tB\x05\x92M\x02\x08\x01R\x0cinplaceField\x12\x45\n\x1brecreate_suppressable_field\x18\x02 \x01(\tB\x05\x92M\x02\x08\x06R\x19recreateSuppressableField\x12\x61\n*recreate_suppressable_restart_stream_field\x18\x04 \x01(\tB\x05\x92M\x02\x08\x08R&recreateSuppressableRestartStreamField\x12+\n\x11unannotated_field\x18\x03 \x01(\tR\x10unannotatedField\"Q\n\x13\x44iffTestNestedOuter\x12:\n\x04\x61rgs\x18\x01 \x01(\x0b\x32\x1f.tecton_proto.args.DiffTestArgsB\x05\x92M\x02\x08\x01R\x04\x61rgsB\x13\n\x0f\x63om.tecton.argsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.diff_test__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001'
  _DIFFTESTARGS.fields_by_name['new_field']._options = None
  _DIFFTESTARGS.fields_by_name['new_field']._serialized_options = b'\222M\002\010\003'
  _DIFFTESTARGS.fields_by_name['old_field']._options = None
  _DIFFTESTARGS.fields_by_name['old_field']._serialized_options = b'\222M\002\010\004'
  _DIFFTESTARGS.fields_by_name['passive_field']._options = None
  _DIFFTESTARGS.fields_by_name['passive_field']._serialized_options = b'\222M\002\010\005'
  _DIFFTESTARGS.fields_by_name['recreate_suppressable_field']._options = None
  _DIFFTESTARGS.fields_by_name['recreate_suppressable_field']._serialized_options = b'\222M\002\010\006'
  _DIFFTESTARGS.fields_by_name['recreate_suppressable_nested_field']._options = None
  _DIFFTESTARGS.fields_by_name['recreate_suppressable_nested_field']._serialized_options = b'\222M\002\010\006'
  _DIFFTESTARGS.fields_by_name['recreate_suppressable_invalidate_checkpoints_field']._options = None
  _DIFFTESTARGS.fields_by_name['recreate_suppressable_invalidate_checkpoints_field']._serialized_options = b'\222M\002\010\007'
  _DIFFTESTARGS.fields_by_name['recreate_suppressable_invalidate_checkpoints_nested_field']._options = None
  _DIFFTESTARGS.fields_by_name['recreate_suppressable_invalidate_checkpoints_nested_field']._serialized_options = b'\222M\002\010\007'
  _DIFFTESTARGS.fields_by_name['recreate_suppressable_restart_stream_field']._options = None
  _DIFFTESTARGS.fields_by_name['recreate_suppressable_restart_stream_field']._serialized_options = b'\222M\002\010\010'
  _DIFFTESTARGS.fields_by_name['renamed_primitive_field']._options = None
  _DIFFTESTARGS.fields_by_name['renamed_primitive_field']._serialized_options = b'\222M\032*\030\n\026former_primitive_field\222M\t*\007\022\0050.6.0'
  _DIFFTESTARGS.fields_by_name['renamed_message_field']._options = None
  _DIFFTESTARGS.fields_by_name['renamed_message_field']._serialized_options = b'\222M\030*\026\n\024former_message_field\222M\t*\007\022\0050.6.0'
  _DIFFTESTNESTEDINNER.fields_by_name['inplace_field']._options = None
  _DIFFTESTNESTEDINNER.fields_by_name['inplace_field']._serialized_options = b'\222M\002\010\001'
  _DIFFTESTNESTEDINNER.fields_by_name['recreate_suppressable_field']._options = None
  _DIFFTESTNESTEDINNER.fields_by_name['recreate_suppressable_field']._serialized_options = b'\222M\002\010\006'
  _DIFFTESTNESTEDINNER.fields_by_name['recreate_suppressable_restart_stream_field']._options = None
  _DIFFTESTNESTEDINNER.fields_by_name['recreate_suppressable_restart_stream_field']._serialized_options = b'\222M\002\010\010'
  _DIFFTESTNESTEDOUTER.fields_by_name['args']._options = None
  _DIFFTESTNESTEDOUTER.fields_by_name['args']._serialized_options = b'\222M\002\010\001'
  _DIFFTESTFOO._serialized_start=237
  _DIFFTESTFOO._serialized_end=300
  _DIFFTESTARGS._serialized_start=303
  _DIFFTESTARGS._serialized_end=1536
  _DIFFTESTNESTEDINNER._serialized_start=1539
  _DIFFTESTNESTEDINNER._serialized_end=1819
  _DIFFTESTNESTEDOUTER._serialized_start=1821
  _DIFFTESTNESTEDOUTER._serialized_end=1902
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import functools
import warnings
from inspect import getfullargspec

from pyspark.rdd import PythonEvalType
from pyspark.sql.pandas.typehints import infer_eval_type
from pyspark.sql.pandas.utils import require_minimum_pandas_version, require_minimum_pyarrow_version
from pyspark.sql.types import DataType
from pyspark.sql.udf import _create_udf


class PandasUDFType(object):
    """Pandas UDF Types. See :meth:`pyspark.sql.functions.pandas_udf`.
    """
    SCALAR = PythonEvalType.SQL_SCALAR_PANDAS_UDF

    SCALAR_ITER = PythonEvalType.SQL_SCALAR_PANDAS_ITER_UDF

    GROUPED_MAP = PythonEvalType.SQL_GROUPED_MAP_PANDAS_UDF

    GROUPED_AGG = PythonEvalType.SQL_GROUPED_AGG_PANDAS_UDF


def pandas_udf(f=None, returnType=None, functionType=None):
    """
    Creates a pandas user defined function (a.k.a. vectorized user defined function).

    Pandas UDFs are user defined functions that are executed by Spark using Arrow to transfer
    data and Pandas to work with the data, which allows vectorized operations. A Pandas UDF
    is defined using the `pandas_udf` as a decorator or to wrap the function, and no
    additional configuration is required. A Pandas UDF behaves as a regular PySpark function
    API in general.

    .. versionadded:: 2.3.0

    Parameters
    ----------
    f : function, optional
        user-defined function. A python function if used as a standalone function
    returnType : :class:`pyspark.sql.types.DataType` or str, optional
        the return type of the user-defined function. The value can be either a
        :class:`pyspark.sql.types.DataType` object or a DDL-formatted type string.
    functionType : int, optional
        an enum value in :class:`pyspark.sql.functions.PandasUDFType`.
        Default: SCALAR. This parameter exists for compatibility.
        Using Python type hints is encouraged.

    Examples
    --------
    In order to use this API, customarily the below are imported:

    >>> import pandas as pd
    >>> from pyspark.sql.functions import pandas_udf

    From Spark 3.0 with Python 3.6+, `Python type hints <https://www.python.org/dev/peps/pep-0484>`_
    detect the function types as below:

    >>> @pandas_udf(IntegerType())
    ... def slen(s: pd.Series) -> pd.Series:
    ...     return s.str.len()

    Prior to Spark 3.0, the pandas UDF used `functionType` to decide the execution type as below:

    >>> from pyspark.sql.functions import PandasUDFType
    >>> from pyspark.sql.types import IntegerType
    >>> @pandas_udf(IntegerType(), PandasUDFType.SCALAR)
    ... def slen(s):
    ...     return s.str.len()

    It is preferred to specify type hints for the pandas UDF instead of specifying pandas UDF
    type via `functionType` which will be deprecated in the future releases.

    Note that the type hint should use `pandas.Series` in all cases but there is one variant
    that `pandas.DataFrame` should be used for its input or output type hint instead when the input
    or output column is of :class:`pyspark.sql.types.StructType`. The following example shows
    a Pandas UDF which takes long column, string column and struct column, and outputs a struct
    column. It requires the function to specify the type hints of `pandas.Series` and
    `pandas.DataFrame` as below:

    >>> @pandas_udf("col1 string, col2 long")
    >>> def func(s1: pd.Series, s2: pd.Series, s3: pd.DataFrame) -> pd.DataFrame:
    ...     s3['col2'] = s1 + s2.str.len()
    ...     return s3
    ...
    >>> # Create a Spark DataFrame that has three columns including a struct column.
    ... df = spark.createDataFrame(
    ...     [[1, "a string", ("a nested string",)]],
    ...     "long_col long, string_col string, struct_col struct<col1:string>")
    >>> df.printSchema()
    root
    |-- long_column: long (nullable = true)
    |-- string_column: string (nullable = true)
    |-- struct_column: struct (nullable = true)
    |    |-- col1: string (nullable = true)
    >>> df.select(func("long_col", "string_col", "struct_col")).printSchema()
    |-- func(long_col, string_col, struct_col): struct (nullable = true)
    |    |-- col1: string (nullable = true)
    |    |-- col2: long (nullable = true)

    In the following sections, it describes the combinations of the supported type hints. For
    simplicity, `pandas.DataFrame` variant is omitted.

    * Series to Series
        `pandas.Series`, ... -> `pandas.Series`

        The function takes one or more `pandas.Series` and outputs one `pandas.Series`.
        The output of the function should always be of the same length as the input.

        >>> @pandas_udf("string")
        ... def to_upper(s: pd.Series) -> pd.Series:
        ...     return s.str.upper()
        ...
        >>> df = spark.createDataFrame([("John Doe",)], ("name",))
        >>> df.select(to_upper("name")).show()
        +--------------+
        |to_upper(name)|
        +--------------+
        |      JOHN DOE|
        +--------------+

        >>> @pandas_udf("first string, last string")
        ... def split_expand(s: pd.Series) -> pd.DataFrame:
        ...     return s.str.split(expand=True)
        ...
        >>> df = spark.createDataFrame([("John Doe",)], ("name",))
        >>> df.select(split_expand("name")).show()
        +------------------+
        |split_expand(name)|
        +------------------+
        |       [John, Doe]|
        +------------------+

        .. note:: The length of the input is not that of the whole input column, but is the
            length of an internal batch used for each call to the function.

    * Iterator of Series to Iterator of Series
        `Iterator[pandas.Series]` -> `Iterator[pandas.Series]`

        The function takes an iterator of `pandas.Series` and outputs an iterator of
        `pandas.Series`. In this case, the created pandas UDF instance requires one input
        column when this is called as a PySpark column. The length of the entire output from
        the function should be the same length of the entire input; therefore, it can
        prefetch the data from the input iterator as long as the lengths are the same.

        It is also useful when the UDF execution
        requires initializing some states although internally it works identically as
        Series to Series case. The pseudocode below illustrates the example.

        .. highlight:: python
        .. code-block:: python

            @pandas_udf("long")
            def calculate(iterator: Iterator[pd.Series]) -> Iterator[pd.Series]:
                # Do some expensive initialization with a state
                state = very_expensive_initialization()
                for x in iterator:
                    # Use that state for whole iterator.
                    yield calculate_with_state(x, state)

            df.select(calculate("value")).show()

        >>> from typing import Iterator
        >>> @pandas_udf("long")
        ... def plus_one(iterator: Iterator[pd.Series]) -> Iterator[pd.Series]:
        ...     for s in iterator:
        ...         yield s + 1
        ...
        >>> df = spark.createDataFrame(pd.DataFrame([1, 2, 3], columns=["v"]))
        >>> df.select(plus_one(df.v)).show()
        +-----------+
        |plus_one(v)|
        +-----------+
        |          2|
        |          3|
        |          4|
        +-----------+

        .. note:: The length of each series is the length of a batch internally used.

    * Iterator of Multiple Series to Iterator of Series
        `Iterator[Tuple[pandas.Series, ...]]` -> `Iterator[pandas.Series]`

        The function takes an iterator of a tuple of multiple `pandas.Series` and outputs an
        iterator of `pandas.Series`. In this case, the created pandas UDF instance requires
        input columns as many as the series when this is called as a PySpark column.
        Otherwise, it has the same characteristics and restrictions as Iterator of Series
        to Iterator of Series case.

        >>> from typing import Iterator, Tuple
        >>> from pyspark.sql.functions import struct, col
        >>> @pandas_udf("long")
        ... def multiply(iterator: Iterator[Tuple[pd.Series, pd.DataFrame]]) -> Iterator[pd.Series]:
        ...     for s1, df in iterator:
        ...         yield s1 * df.v
        ...
        >>> df = spark.createDataFrame(pd.DataFrame([1, 2, 3], columns=["v"]))
        >>> df.withColumn('output', multiply(col("v"), struct(col("v")))).show()
        +---+------+
        |  v|output|
        +---+------+
        |  1|     1|
        |  2|     4|
        |  3|     9|
        +---+------+

        .. note:: The length of each series is the length of a batch internally used.

    * Series to Scalar
        `pandas.Series`, ... -> `Any`

        The function takes `pandas.Series` and returns a scalar value. The `returnType`
        should be a primitive data type, and the returned scalar can be either a python primitive
        type, e.g., int or float or a numpy data type, e.g., numpy.int64 or numpy.float64.
        `Any` should ideally be a specific scalar type accordingly.

        >>> @pandas_udf("double")
        ... def mean_udf(v: pd.Series) -> float:
        ...     return v.mean()
        ...
        >>> df = spark.createDataFrame(
        ...     [(1, 1.0), (1, 2.0), (2, 3.0), (2, 5.0), (2, 10.0)], ("id", "v"))
        >>> df.groupby("id").agg(mean_udf(df['v'])).show()
        +---+-----------+
        | id|mean_udf(v)|
        +---+-----------+
        |  1|        1.5|
        |  2|        6.0|
        +---+-----------+

        This UDF can also be used as window functions as below:

        >>> from pyspark.sql import Window
        >>> @pandas_udf("double")
        ... def mean_udf(v: pd.Series) -> float:
        ...     return v.mean()
        ...
        >>> df = spark.createDataFrame(
        ...     [(1, 1.0), (1, 2.0), (2, 3.0), (2, 5.0), (2, 10.0)], ("id", "v"))
        >>> w = Window.partitionBy('id').orderBy('v').rowsBetween(-1, 0)
        >>> df.withColumn('mean_v', mean_udf("v").over(w)).show()
        +---+----+------+
        | id|   v|mean_v|
        +---+----+------+
        |  1| 1.0|   1.0|
        |  1| 2.0|   1.5|
        |  2| 3.0|   3.0|
        |  2| 5.0|   4.0|
        |  2|10.0|   7.5|
        +---+----+------+

        .. note:: For performance reasons, the input series to window functions are not copied.
            Therefore, mutating the input series is not allowed and will cause incorrect results.
            For the same reason, users should also not rely on the index of the input series.

    Notes
    -----
    The user-defined functions do not support conditional expressions or short circuiting
    in boolean expressions and it ends up with being executed all internally. If the functions
    can fail on special rows, the workaround is to incorporate the condition into the functions.

    The user-defined functions do not take keyword arguments on the calling side.

    The data type of returned `pandas.Series` from the user-defined functions should be
    matched with defined `returnType` (see :meth:`types.to_arrow_type` and
    :meth:`types.from_arrow_type`). When there is mismatch between them, Spark might do
    conversion on returned data. The conversion is not guaranteed to be correct and results
    should be checked for accuracy by users.

    Currently,
    :class:`pyspark.sql.types.ArrayType` of :class:`pyspark.sql.types.TimestampType` and
    nested :class:`pyspark.sql.types.StructType`
    are currently not supported as output types.

    See Also
    --------
    pyspark.sql.GroupedData.agg
    pyspark.sql.DataFrame.mapInPandas
    pyspark.sql.GroupedData.applyInPandas
    pyspark.sql.PandasCogroupedOps.applyInPandas
    pyspark.sql.UDFRegistration.register
    """

    # The following table shows most of Pandas data and SQL type conversions in Pandas UDFs that
    # are not yet visible to the user. Some of behaviors are buggy and might be changed in the near
    # future. The table might have to be eventually documented externally.
    # Please see SPARK-28132's PR to see the codes in order to generate the table below.
    #
    # +-----------------------------+----------------------+------------------+------------------+------------------+--------------------+--------------------+------------------+------------------+------------------+------------------+--------------+--------------+--------------+-----------------------------------+-----------------------------------------------------+-----------------+--------------------+-----------------------------+--------------+-----------------+------------------+---------------+--------------------------------+  # noqa
    # |SQL Type \ Pandas Value(Type)|None(object(NoneType))|        True(bool)|           1(int8)|          1(int16)|            1(int32)|            1(int64)|          1(uint8)|         1(uint16)|         1(uint32)|         1(uint64)|  1.0(float16)|  1.0(float32)|  1.0(float64)|1970-01-01 00:00:00(datetime64[ns])|1970-01-01 00:00:00-05:00(datetime64[ns, US/Eastern])|a(object(string))|  1(object(Decimal))|[1 2 3](object(array[int32]))| 1.0(float128)|(1+0j)(complex64)|(1+0j)(complex128)|    A(category)|1 days 00:00:00(timedelta64[ns])|  # noqa
    # +-----------------------------+----------------------+------------------+------------------+------------------+--------------------+--------------------+------------------+------------------+------------------+------------------+--------------+--------------+--------------+-----------------------------------+-----------------------------------------------------+-----------------+--------------------+-----------------------------+--------------+-----------------+------------------+---------------+--------------------------------+  # noqa
    # |                      boolean|                  None|              True|              True|              True|                True|                True|              True|              True|              True|              True|          True|          True|          True|                                  X|                                                    X|                X|                   X|                            X|             X|                X|                 X|              X|                               X|  # noqa
    # |                      tinyint|                  None|                 1|                 1|                 1|                   1|                   1|                 1|                 1|                 1|                 1|             1|             1|             1|                                  X|                                                    X|                X|                   1|                            X|             X|                X|                 X|              X|                               X|  # noqa
    # |                     smallint|                  None|                 1|                 1|                 1|                   1|                   1|                 1|                 1|                 1|                 1|             1|             1|             1|                                  X|                                                    X|                X|                   1|                            X|             X|                X|                 X|              X|                               X|  # noqa
    # |                          int|                  None|                 1|                 1|                 1|                   1|                   1|                 1|                 1|                 1|                 1|             1|             1|             1|                                  X|                                                    X|                X|                   1|                            X|             X|                X|                 X|              X|                               X|  # noqa
    # |                       bigint|                  None|                 1|                 1|                 1|                   1|                   1|                 1|                 1|                 1|                 1|             1|             1|             1|                                  0|                                       18000000000000|                X|                   1|                            X|             X|                X|                 X|              X|                  86400000000000|  # noqa
    # |                        float|                  None|               1.0|               1.0|               1.0|                 1.0|                 1.0|               1.0|               1.0|               1.0|               1.0|           1.0|           1.0|           1.0|                                  X|                                                    X|                X|                   X|                            X|             X|                X|                 X|              X|                               X|  # noqa
    # |                       double|                  None|               1.0|               1.0|               1.0|                 1.0|                 1.0|               1.0|               1.0|               1.0|               1.0|           1.0|           1.0|           1.0|                                  X|                                                    X|                X|                   X|                            X|             X|                X|                 X|              X|                               X|  # noqa
    # |                         date|                  None|                 X|                 X|                 X|datetime.date(197...|                   X|                 X|                 X|                 X|                 X|             X|             X|             X|               datetime.date(197...|                                 datetime.date(197...|                X|datetime.date(197...|                            X|             X|                X|                 X|              X|                               X|  # noqa
    # |                    timestamp|                  None|                 X|                 X|                 X|                   X|datetime.datetime...|                 X|                 X|                 X|                 X|             X|             X|             X|               datetime.datetime...|                                 datetime.datetime...|                X|datetime.datetime...|                            X|             X|                X|                 X|              X|                               X|  # noqa
    # |                       string|                  None|                 X|                 X|                 X|                   X|                   X|                 X|                 X|                 X|                 X|             X|             X|             X|                                  X|                                                    X|              'a'|                   X|                            X|             X|                X|                 X|            'A'|                               X|  # noqa
    # |                decimal(10,0)|                  None|                 X|                 X|                 X|                   X|                   X|                 X|                 X|                 X|                 X|             X|             X|             X|                                  X|                                                    X|                X|        Decimal('1')|                            X|             X|                X|                 X|              X|                               X|  # noqa
    # |                   array<int>|                  None|                 X|                 X|                 X|                   X|                   X|                 X|                 X|                 X|                 X|             X|             X|             X|                                  X|                                                    X|                X|                   X|                    [1, 2, 3]|             X|                X|                 X|              X|                               X|  # noqa
    # |              map<string,int>|                     X|                 X|                 X|                 X|                   X|                   X|                 X|                 X|                 X|                 X|             X|             X|             X|                                  X|                                                    X|                X|                   X|                            X|             X|                X|                 X|              X|                               X|  # noqa
    # |               struct<_1:int>|                     X|                 X|                 X|                 X|                   X|                   X|                 X|                 X|                 X|                 X|             X|             X|             X|                                  X|                                                    X|                X|                   X|                            X|             X|                X|                 X|              X|                               X|  # noqa
    # |                       binary|                  None|bytearray(b'\x01')|bytearray(b'\x01')|bytearray(b'\x01')|  bytearray(b'\x01')|  bytearray(b'\x01')|bytearray(b'\x01')|bytearray(b'\x01')|bytearray(b'\x01')|bytearray(b'\x01')|bytearray(b'')|bytearray(b'')|bytearray(b'')|                     bytearray(b'')|                                       bytearray(b'')|  bytearray(b'a')|                   X|                            X|bytearray(b'')|   bytearray(b'')|    bytearray(b'')|bytearray(b'A')|                  bytearray(b'')|  # noqa
    # +-----------------------------+----------------------+------------------+------------------+------------------+--------------------+--------------------+------------------+------------------+------------------+------------------+--------------+--------------+--------------+-----------------------------------+-----------------------------------------------------+-----------------+--------------------+-----------------------------+--------------+-----------------+------------------+---------------+--------------------------------+  # noqa    #
    #
    # Note: DDL formatted string is used for 'SQL Type' for simplicity. This string can be
    #       used in `returnType`.
    # Note: The values inside of the table are generated by `repr`.
    # Note: Python 3.7.3, Pandas 1.1.1 and PyArrow 1.0.1 are used.
    # Note: Timezone is KST.
    # Note: 'X' means it throws an exception during the conversion.
    require_minimum_pandas_version()
    require_minimum_pyarrow_version()

    # decorator @pandas_udf(returnType, functionType)
    is_decorator = f is None or isinstance(f, (str, DataType))

    if is_decorator:
        # If DataType has been passed as a positional argument
        # for decorator use it as a returnType
        return_type = f or returnType

        if functionType is not None:
            # @pandas_udf(dataType, functionType=functionType)
            # @pandas_udf(returnType=dataType, functionType=functionType)
            eval_type = functionType
        elif returnType is not None and isinstance(returnType, int):
            # @pandas_udf(dataType, functionType)
            eval_type = returnType
        else:
            # @pandas_udf(dataType) or @pandas_udf(returnType=dataType)
            eval_type = None
    else:
        return_type = returnType

        if functionType is not None:
            eval_type = functionType
        else:
            eval_type = None

    if return_type is None:
        raise ValueError("Invalid return type: returnType can not be None")

    if eval_type not in [PythonEvalType.SQL_SCALAR_PANDAS_UDF,
                         PythonEvalType.SQL_SCALAR_PANDAS_ITER_UDF,
                         PythonEvalType.SQL_GROUPED_MAP_PANDAS_UDF,
                         PythonEvalType.SQL_GROUPED_AGG_PANDAS_UDF,
                         PythonEvalType.SQL_MAP_PANDAS_ITER_UDF,
                         PythonEvalType.SQL_COGROUPED_MAP_PANDAS_UDF,
                         None]:  # None means it should infer the type from type hints.

        raise ValueError("Invalid function type: "
                         "functionType must be one the values from PandasUDFType")

    if is_decorator:
        return functools.partial(_create_pandas_udf, returnType=return_type, evalType=eval_type)
    else:
        return _create_pandas_udf(f=f, returnType=return_type, evalType=eval_type)


def _create_pandas_udf(f, returnType, evalType):
    argspec = getfullargspec(f)

    # pandas UDF by type hints.
    from inspect import signature

    if evalType in [PythonEvalType.SQL_SCALAR_PANDAS_UDF,
                    PythonEvalType.SQL_SCALAR_PANDAS_ITER_UDF,
                    PythonEvalType.SQL_GROUPED_AGG_PANDAS_UDF]:
        warnings.warn(
            "In Python 3.6+ and Spark 3.0+, it is preferred to specify type hints for "
            "pandas UDF instead of specifying pandas UDF type which will be deprecated "
            "in the future releases. See SPARK-28264 for more details.", UserWarning)
    elif evalType in [PythonEvalType.SQL_GROUPED_MAP_PANDAS_UDF,
                      PythonEvalType.SQL_MAP_PANDAS_ITER_UDF,
                      PythonEvalType.SQL_COGROUPED_MAP_PANDAS_UDF]:
        # In case of 'SQL_GROUPED_MAP_PANDAS_UDF',  deprecation warning is being triggered
        # at `apply` instead.
        # In case of 'SQL_MAP_PANDAS_ITER_UDF' and 'SQL_COGROUPED_MAP_PANDAS_UDF', the
        # evaluation type will always be set.
        pass
    elif len(argspec.annotations) > 0:
        evalType = infer_eval_type(signature(f))
        assert evalType is not None

    if evalType is None:
        # Set default is scalar UDF.
        evalType = PythonEvalType.SQL_SCALAR_PANDAS_UDF

    if (evalType == PythonEvalType.SQL_SCALAR_PANDAS_UDF or
            evalType == PythonEvalType.SQL_SCALAR_PANDAS_ITER_UDF) and \
            len(argspec.args) == 0 and \
            argspec.varargs is None:
        raise ValueError(
            "Invalid function: 0-arg pandas_udfs are not supported. "
            "Instead, create a 1-arg pandas_udf and ignore the arg in your function."
        )

    if evalType == PythonEvalType.SQL_GROUPED_MAP_PANDAS_UDF \
            and len(argspec.args) not in (1, 2):
        raise ValueError(
            "Invalid function: pandas_udf with function type GROUPED_MAP or "
            "the function in groupby.applyInPandas "
            "must take either one argument (data) or two arguments (key, data).")

    if evalType == PythonEvalType.SQL_COGROUPED_MAP_PANDAS_UDF \
            and len(argspec.args) not in (2, 3):
        raise ValueError(
            "Invalid function: the function in cogroup.applyInPandas "
            "must take either two arguments (left, right) "
            "or three arguments (key, left, right).")

    return _create_udf(f, returnType, evalType)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/api/featureservice/feature_service__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.api.featureservice import feature_service_request__client_pb2 as tecton__proto_dot_api_dot_featureservice_dot_feature__service__request____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n=tecton_proto/api/featureservice/feature_service__client.proto\x12 tecton_proto.api.feature_service\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x45tecton_proto/api/featureservice/feature_service_request__client.proto\"\xcf\x01\n\x0bNameAndType\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12[\n\tdata_type\x18\x03 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\x08\x64\x61taType\x12O\n\x04type\x18\x02 \x01(\x0e\x32\x37.tecton_proto.api.feature_service.FeatureServerDataTypeB\x02\x18\x01R\x04type\"\x94\x01\n\x11GetFeaturesResult\x12\x32\n\x08\x66\x65\x61tures\x18\x05 \x03(\x0b\x32\x16.google.protobuf.ValueR\x08\x66\x65\x61tures\x12\x33\n\tjoin_keys\x18\x06 \x03(\x0b\x32\x16.google.protobuf.ValueR\x08joinKeysJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05\"\xdc\x03\n\x1c\x46\x65\x61tureServerComplexDataType\x12K\n\x04type\x18\x01 \x01(\x0e\x32\x37.tecton_proto.api.feature_service.FeatureServerDataTypeR\x04type\x12\x61\n\x0c\x65lement_type\x18\x02 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\x0b\x65lementType\x12R\n\x06\x66ields\x18\x03 \x03(\x0b\x32:.tecton_proto.api.feature_service.FeatureServerStructFieldR\x06\x66ields\x12Y\n\x08key_type\x18\x04 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\x07keyType\x12]\n\nvalue_type\x18\x05 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\tvalueType\"\x8b\x01\n\x18\x46\x65\x61tureServerStructField\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12[\n\tdata_type\x18\x02 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\x08\x64\x61taType\"\xdf\x03\n\x07SloInfo\x12!\n\x0cslo_eligible\x18\x01 \x01(\x08R\x0bsloEligible\x12\x35\n\x17slo_server_time_seconds\x18\x05 \x01(\x01R\x14sloServerTimeSeconds\x12t\n\x19slo_ineligibility_reasons\x18\x02 \x03(\x0e\x32\x38.tecton_proto.api.feature_service.SloIneligibilityReasonR\x17sloIneligibilityReasons\x12?\n\x1c\x64ynamodb_response_size_bytes\x18\x03 \x01(\x05R\x19\x64ynamodbResponseSizeBytes\x12.\n\x13server_time_seconds\x18\x04 \x01(\x01R\x11serverTimeSeconds\x12,\n\x12store_time_seconds\x18\x06 \x01(\x01R\x10storeTimeSeconds\x12*\n\x11store_max_latency\x18\x07 \x01(\x01R\x0fstoreMaxLatency\x12\x39\n\x19store_response_size_bytes\x18\x08 \x01(\x05R\x16storeResponseSizeBytes\"\xea\x04\n\x08Metadata\x12K\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32/.tecton_proto.api.feature_service.Metadata.ItemR\x08\x66\x65\x61tures\x12L\n\tjoin_keys\x18\x02 \x03(\x0b\x32/.tecton_proto.api.feature_service.Metadata.ItemR\x08joinKeys\x12\x44\n\x08slo_info\x18\x03 \x01(\x0b\x32).tecton_proto.api.feature_service.SloInfoR\x07sloInfo\x12\'\n\x0fpartial_results\x18\x04 \x01(\x08R\x0epartialResults\x1a\xd3\x02\n\x04Item\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x41\n\x0e\x65\x66\x66\x65\x63tive_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\reffectiveTime\x12O\n\x04type\x18\x03 \x01(\x0e\x32\x37.tecton_proto.api.feature_service.FeatureServerDataTypeB\x02\x18\x01R\x04type\x12[\n\tdata_type\x18\x04 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\x08\x64\x61taType\x12@\n\x06status\x18\x06 \x01(\x0e\x32(.tecton_proto.api.feature_service.StatusR\x06statusJ\x04\x08\x05\x10\x06\"\xaa\x01\n\x13GetFeaturesResponse\x12K\n\x06result\x18\x02 \x01(\x0b\x32\x33.tecton_proto.api.feature_service.GetFeaturesResultR\x06result\x12\x46\n\x08metadata\x18\x03 \x01(\x0b\x32*.tecton_proto.api.feature_service.MetadataR\x08metadata\"\xba\x01\n\x18GetFeaturesBatchResponse\x12K\n\x06result\x18\x02 \x03(\x0b\x32\x33.tecton_proto.api.feature_service.GetFeaturesResultR\x06result\x12K\n\x08metadata\x18\x04 \x01(\x0b\x32/.tecton_proto.api.feature_service.BatchMetadataR\x08metadataJ\x04\x08\x03\x10\x04\"\xba\x02\n\x0c\x42\x61tchSloInfo\x12!\n\x0cslo_eligible\x18\x01 \x01(\x08R\x0bsloEligible\x12\x35\n\x17slo_server_time_seconds\x18\x02 \x01(\x01R\x14sloServerTimeSeconds\x12t\n\x19slo_ineligibility_reasons\x18\x03 \x03(\x0e\x32\x38.tecton_proto.api.feature_service.SloIneligibilityReasonR\x17sloIneligibilityReasons\x12.\n\x13server_time_seconds\x18\x04 \x01(\x01R\x11serverTimeSeconds\x12*\n\x11store_max_latency\x18\x05 \x01(\x01R\x0fstoreMaxLatency\"\xd5\x04\n\rBatchMetadata\x12P\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32\x34.tecton_proto.api.feature_service.BatchMetadata.ItemR\x08\x66\x65\x61tures\x12Q\n\tjoin_keys\x18\x02 \x03(\x0b\x32\x34.tecton_proto.api.feature_service.BatchMetadata.ItemR\x08joinKeys\x12\x44\n\x08slo_info\x18\x03 \x03(\x0b\x32).tecton_proto.api.feature_service.SloInfoR\x07sloInfo\x12T\n\x0e\x62\x61tch_slo_info\x18\x04 \x01(\x0b\x32..tecton_proto.api.feature_service.BatchSloInfoR\x0c\x62\x61tchSloInfo\x1a\x82\x02\n\x04Item\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x41\n\x0e\x65\x66\x66\x65\x63tive_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\reffectiveTime\x12[\n\tdata_type\x18\x03 \x01(\x0b\x32>.tecton_proto.api.feature_service.FeatureServerComplexDataTypeR\x08\x64\x61taType\x12@\n\x06status\x18\x05 \x03(\x0e\x32(.tecton_proto.api.feature_service.StatusR\x06statusJ\x04\x08\x04\x10\x05\"i\n\x16ServiceMetadataRequest\x12O\n\x06params\x18\x01 \x01(\x0b\x32\x37.tecton_proto.api.feature_service.FeatureServiceLocatorR\x06params\"\xf3\x03\n\x17ServiceMetadataResponse\x12\x66\n\x14\x66\x65\x61ture_service_type\x18\x01 \x01(\x0e\x32\x34.tecton_proto.api.feature_service.FeatureServiceTypeR\x12\x66\x65\x61tureServiceType\x12U\n\x0finput_join_keys\x18\x02 \x03(\x0b\x32-.tecton_proto.api.feature_service.NameAndTypeR\rinputJoinKeys\x12j\n\x1ainput_request_context_keys\x18\x03 \x03(\x0b\x32-.tecton_proto.api.feature_service.NameAndTypeR\x17inputRequestContextKeys\x12W\n\x10output_join_keys\x18\x04 \x03(\x0b\x32-.tecton_proto.api.feature_service.NameAndTypeR\x0eoutputJoinKeys\x12T\n\x0e\x66\x65\x61ture_values\x18\x05 \x03(\x0b\x32-.tecton_proto.api.feature_service.NameAndTypeR\rfeatureValues\"g\n\x14QueryFeaturesRequest\x12O\n\x06params\x18\x01 \x01(\x0b\x32\x37.tecton_proto.api.feature_service.GetFeaturesParametersR\x06params\"\xae\x01\n\x15QueryFeaturesResponse\x12M\n\x07results\x18\x01 \x03(\x0b\x32\x33.tecton_proto.api.feature_service.GetFeaturesResultR\x07results\x12\x46\n\x08metadata\x18\x02 \x01(\x0b\x32*.tecton_proto.api.feature_service.MetadataR\x08metadata\"\xe9\x02\n\x15\x46\x65\x61tureServiceLocator\x12.\n\x12\x66\x65\x61ture_service_id\x18\x01 \x01(\tH\x00R\x10\x66\x65\x61tureServiceId\x12\x32\n\x14\x66\x65\x61ture_service_name\x18\x02 \x01(\tH\x00R\x12\x66\x65\x61tureServiceName\x12.\n\x12\x66\x65\x61ture_package_id\x18\x04 \x01(\tH\x00R\x10\x66\x65\x61turePackageId\x12\x32\n\x14\x66\x65\x61ture_package_name\x18\x05 \x01(\tH\x00R\x12\x66\x65\x61turePackageName\x12(\n\x0f\x66\x65\x61ture_view_id\x18\x06 \x01(\tH\x00R\rfeatureViewId\x12,\n\x11\x66\x65\x61ture_view_name\x18\x07 \x01(\tH\x00R\x0f\x66\x65\x61tureViewName\x12%\n\x0eworkspace_name\x18\x03 \x01(\tR\rworkspaceNameB\t\n\x07locator\"\x90\x01\n\x1dGetFeatureServiceStateRequest\x12o\n\x17\x66\x65\x61ture_service_locator\x18\x01 \x01(\x0b\x32\x37.tecton_proto.api.feature_service.FeatureServiceLocatorR\x15\x66\x65\x61tureServiceLocator\"&\n\x1eGetFeatureServiceStateResponseJ\x04\x08\x01\x10\x02*/\n\x12\x46\x65\x61tureServiceType\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\x0c\n\x08WILDCARD\x10\x01*\xa8\x01\n\x15\x46\x65\x61tureServerDataType\x12\x10\n\x0cmissing_type\x10\x00\x12\x0b\n\x07\x62oolean\x10\x01\x12\x0b\n\x07\x66loat64\x10\x02\x12\t\n\x05int64\x10\x03\x12\n\n\x06string\x10\x04\x12\x10\n\x0cstring_array\x10\x05\x12\t\n\x05\x61rray\x10\x06\x12\x0b\n\x07\x66loat32\x10\x07\x12\n\n\x06struct\x10\x08\x12\x07\n\x03map\x10\t\x12\r\n\ttimestamp\x10\n*|\n\x06Status\x12\x0b\n\x07PRESENT\x10\x00\x12\x10\n\x0cMISSING_DATA\x10\x01\x12\x12\n\x0eUNKNOWN_STATUS\x10\x02\x12\x12\n\x0e\x43\x41\x43HED_UNKNOWN\x10\x03\x12\x17\n\x13\x43\x41\x43HED_MISSING_DATA\x10\x04\x12\x12\n\x0e\x43\x41\x43HED_PRESENT\x10\x05*\x9a\x01\n\x16SloIneligibilityReason\x12\x0b\n\x07UNKNOWN\x10\x00\x12)\n%DYNAMODB_RESPONSE_SIZE_LIMIT_EXCEEDED\x10\x01\x12&\n\"REDIS_RESPONSE_SIZE_LIMIT_EXCEEDED\x10\x02\x12 \n\x1cREDIS_LATENCY_LIMIT_EXCEEDED\x10\x03\x32\xc2\x05\n\x0e\x46\x65\x61tureService\x12z\n\x0bGetFeatures\x12\x34.tecton_proto.api.feature_service.GetFeaturesRequest\x1a\x35.tecton_proto.api.feature_service.GetFeaturesResponse\x12\x89\x01\n\x10GetFeaturesBatch\x12\x39.tecton_proto.api.feature_service.GetFeaturesBatchRequest\x1a:.tecton_proto.api.feature_service.GetFeaturesBatchResponse\x12\x80\x01\n\rQueryFeatures\x12\x36.tecton_proto.api.feature_service.QueryFeaturesRequest\x1a\x37.tecton_proto.api.feature_service.QueryFeaturesResponse\x12\x9b\x01\n\x16GetFeatureServiceState\x12?.tecton_proto.api.feature_service.GetFeatureServiceStateRequest\x1a@.tecton_proto.api.feature_service.GetFeatureServiceStateResponse\x12\x86\x01\n\x0fServiceMetadata\x12\x38.tecton_proto.api.feature_service.ServiceMetadataRequest\x1a\x39.tecton_proto.api.feature_service.ServiceMetadataResponseB6Z4github.com/tecton-ai/tecton_proto/api/featureservice')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.api.featureservice.feature_service__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z4github.com/tecton-ai/tecton_proto/api/featureservice'
  _NAMEANDTYPE.fields_by_name['type']._options = None
  _NAMEANDTYPE.fields_by_name['type']._serialized_options = b'\030\001'
  _METADATA_ITEM.fields_by_name['type']._options = None
  _METADATA_ITEM.fields_by_name['type']._serialized_options = b'\030\001'
  _FEATURESERVICETYPE._serialized_start=5039
  _FEATURESERVICETYPE._serialized_end=5086
  _FEATURESERVERDATATYPE._serialized_start=5089
  _FEATURESERVERDATATYPE._serialized_end=5257
  _STATUS._serialized_start=5259
  _STATUS._serialized_end=5383
  _SLOINELIGIBILITYREASON._serialized_start=5386
  _SLOINELIGIBILITYREASON._serialized_end=5540
  _NAMEANDTYPE._serialized_start=234
  _NAMEANDTYPE._serialized_end=441
  _GETFEATURESRESULT._serialized_start=444
  _GETFEATURESRESULT._serialized_end=592
  _FEATURESERVERCOMPLEXDATATYPE._serialized_start=595
  _FEATURESERVERCOMPLEXDATATYPE._serialized_end=1071
  _FEATURESERVERSTRUCTFIELD._serialized_start=1074
  _FEATURESERVERSTRUCTFIELD._serialized_end=1213
  _SLOINFO._serialized_start=1216
  _SLOINFO._serialized_end=1695
  _METADATA._serialized_start=1698
  _METADATA._serialized_end=2316
  _METADATA_ITEM._serialized_start=1977
  _METADATA_ITEM._serialized_end=2316
  _GETFEATURESRESPONSE._serialized_start=2319
  _GETFEATURESRESPONSE._serialized_end=2489
  _GETFEATURESBATCHRESPONSE._serialized_start=2492
  _GETFEATURESBATCHRESPONSE._serialized_end=2678
  _BATCHSLOINFO._serialized_start=2681
  _BATCHSLOINFO._serialized_end=2995
  _BATCHMETADATA._serialized_start=2998
  _BATCHMETADATA._serialized_end=3595
  _BATCHMETADATA_ITEM._serialized_start=3337
  _BATCHMETADATA_ITEM._serialized_end=3595
  _SERVICEMETADATAREQUEST._serialized_start=3597
  _SERVICEMETADATAREQUEST._serialized_end=3702
  _SERVICEMETADATARESPONSE._serialized_start=3705
  _SERVICEMETADATARESPONSE._serialized_end=4204
  _QUERYFEATURESREQUEST._serialized_start=4206
  _QUERYFEATURESREQUEST._serialized_end=4309
  _QUERYFEATURESRESPONSE._serialized_start=4312
  _QUERYFEATURESRESPONSE._serialized_end=4486
  _FEATURESERVICELOCATOR._serialized_start=4489
  _FEATURESERVICELOCATOR._serialized_end=4850
  _GETFEATURESERVICESTATEREQUEST._serialized_start=4853
  _GETFEATURESERVICESTATEREQUEST._serialized_end=4997
  _GETFEATURESERVICESTATERESPONSE._serialized_start=4999
  _GETFEATURESERVICESTATERESPONSE._serialized_end=5037
  _FEATURESERVICE._serialized_start=5543
  _FEATURESERVICE._serialized_end=6249
# @@protoc_insertion_point(module_scope)
